(MODULE AMA (WITH QOBISCHEME) (MAIN MAIN))
;;; LaHaShem HaAretz U'Mloah

(include "QobiScheme.sch")
(include "hlib-span-basic.sc")

(define *howard-directory* "./data-m19nov2001/")

(set! *program* "ama")
(set! *panic?* #t)

;;; Structures

(define-structure ama:variable t1 t2 count)

(define *event-types*
 '("pick-up" "put-down" "stack" "unstack" "move" "assemble" "disassemble"))

(define *conditions* '("left" "right"))

(define *indices*
 '("00" "01" "02" "03" "04" "05" "06" "07" "08" "09" "10" "11" "12" "13" "14"))

(define (num->condition-index n)
  (if (< n (length *indices*))
      (list "left" (list-ref *indices* n))
      (list "right" (list-ref *indices* (- n 15)))))

(define *num-occurrences* 
  '(("pick-up" 150)
    ("put-down" 150)
    ("stack" 60)
    ("unstack" 60)
    ("move" 30)
    ("assemble" 30)
    ("disassemble" 30)))

(define *templates*
 '((("pick-up" "left") (pick-up moving red green))
   (("pick-up" "right") (pick-up moving red green))
   (("put-down" "left") (put-down moving red green))
   (("put-down" "right") (put-down moving red green))
   (("stack" "left") (stack moving red green blue))
   (("stack" "right") (stack moving red green blue))
   (("unstack" "left") (unstack moving red green blue))
   (("unstack" "right") (unstack moving red green blue))
   (("move" "left") (move moving red green blue))
   (("move" "right") (move moving red green blue))
   (("assemble" "left") (assemble moving red blue green))
   (("assemble" "right") (assemble moving red blue green))
   (("disassemble" "left") (disassemble moving red blue green))
   (("disassemble" "right") (disassemble moving red blue green))))

(define *correct-results*
 '((("pick-up" "left") (pick-up moving red green))
   (("pick-up" "right") (pick-up moving red green))
   (("put-down" "left") (put-down moving red green))
   (("put-down" "right") (put-down moving red green))
   (("stack" "left") (stack moving red green blue)
		     (put-down moving red green))
   (("stack" "right") (stack moving red green blue)
		      (put-down moving red green))
   (("unstack" "left") (unstack moving red green blue)
		       (pick-up moving red green))
   (("unstack" "right") (unstack moving red green blue)
			(pick-up moving red green))
   (("move" "left") (move moving red green blue)
		    (pick-up moving red green)
		    (put-down moving red blue))
   (("move" "right") (move moving red blue green)
		     (pick-up moving red blue)
		     (put-down moving red green))
   (("assemble" "left") (assemble moving red blue green)
			(put-down moving blue green)
			(stack moving red blue green)
			(put-down moving red blue))
   (("assemble" "right") (assemble moving red blue green)
			 (put-down moving blue green)
			 (stack moving red blue green)
			 (put-down moving red blue))
   (("disassemble" "left") (disassemble moving red blue green)
			   (unstack moving red blue green)
			   (pick-up moving red blue)
			   (pick-up moving blue green))
   (("disassemble" "right") (disassemble moving red blue green)
			    (unstack moving red blue green)
			    (pick-up moving red blue)
			    (pick-up moving blue green))))

(define *correct-diagonal-results*
 '((("pick-up" "left") (pick-up moving red green))
   (("pick-up" "right") (pick-up moving red green))
   (("put-down" "left") (put-down moving red green))
   (("put-down" "right") (put-down moving red green))
   (("stack" "left") (stack moving red green blue))
   (("stack" "right") (stack moving red green blue))
   (("unstack" "left") (unstack moving red green blue))
   (("unstack" "right") (unstack moving red green blue))
   (("move" "left") (move moving red green blue))
   (("move" "right") (move moving red blue green))
   (("assemble" "left") (assemble moving red blue green))
   (("assemble" "right") (assemble moving red blue green))
   (("disassemble" "left") (disassemble moving red blue green))
   (("disassemble" "right") (disassemble moving red blue green))))


;;; Variables

(define *lifted?* #f)

(define *variables* '())

(define *variable-counter* 0)

;;; Procedures

(define (first-n n l) (if (or (not n) (< (length l) n)) l (sublist l 0 n)))

(define (minimal-values <? <=? thunk)
 ;; belongs in QobiScheme: except that this should be a macro
 ;; note: There is a bug in Scheme->C which doesn't allow < and <= to shadow
 ;;       the global bindings which is why these are named <? and <=?.
 (let ((minimal-values '()))
  (for-effects
   (let ((value (thunk)))
    (unless (some (lambda (minimal-value) (<=? minimal-value value))
		  minimal-values)
     (set! minimal-values
	   (cons value
		 (remove-if (lambda (minimal-value) (<? value minimal-value))
			    minimal-values))))))
  minimal-values))

(define (minimal-members <? <=? l)
 ;; belongs in QobiScheme: This should replace MINIMAL-MEMBERSP.
 ;; note: There is a bug in Scheme->C which doesn't allow < and <= to shadow
 ;;       the global bindings which is why these are named <? and <=?.
 (let ((minimal-members '()))
  (for-each
   (lambda (x)
    (unless (some (lambda (minimal-member) (<=? minimal-member x))
		  minimal-members)
     (set! minimal-members
	   (cons x
		 (remove-if (lambda (minimal-member) (<? x minimal-member))
			    minimal-members)))))
   l)
  minimal-members))

(define (map-matrix p . ms)
 ;; belongs in QobiScheme
 (apply map-vector (lambda rows (apply map-vector p rows)) ms))

(define (make-upper-triangular-matrix n . initial-value)
 ;; belongs in QobiScheme
 (if (null? initial-value)
     (make-vector (quotient (* n (- n 1)) 2))
     (make-vector (quotient (* n (- n 1)) 2) (first initial-value))))

(define (upper-triangular-matrix-size m)
 ;; belongs in QobiScheme
 (let ((n (quadratic1 0.5 -0.5 (- (vector-length m)))))
  (unless (integer? n)
   (panic "The length of the vector is inappropriate for a square upper-triangular matrix"))
  (inexact->exact (round n))))

(define (upper-triangular-matrix-index m i j)
 ;; belongs in QobiScheme
 (let ((n (upper-triangular-matrix-size m)))
  (unless (and (>= i 0) (< i n) (>= j 0) (< j n) (< i j))
   (panic "Index out of bounds"))
  (+ (quotient (- (* n (- n 1)) (* (- n i) (- n i 1))) 2) (- j i 1))))

(define (upper-triangular-matrix-ref m i j)
 ;; belongs in QobiScheme
 (vector-ref m (upper-triangular-matrix-index m i j)))

(define (upper-triangular-matrix-set! m i j x)
 ;; belongs in QobiScheme
 (vector-set! m (upper-triangular-matrix-index m i j) x))

(define (map-upper-triangular-matrix p . ms)
 ;; belongs in QobiScheme
 (apply map-vector p ms))

(define (modeq l)
 ;; belongs in QobiScheme
 (first
  (first				;arbitrary
   (minimal-members (lambda (x y) (> (length x) (length y)))
		    (lambda (x y) (>= (length x) (length y)))
		    (equivalence-classesq l)))))

(define (modev l)
 ;; belongs in QobiScheme
 (first
  (first				;arbitrary
   (minimal-members (lambda (x y) (> (length x) (length y)))
		    (lambda (x y) (>= (length x) (length y)))
		    (equivalence-classesv l)))))

(define (ama:mode l)
 ;; belongs in QobiScheme
 (first
  (first				;arbitrary
   (minimal-members (lambda (x y) (> (length x) (length y)))
		    (lambda (x y) (>= (length x) (length y)))
		    (equivalence-classes l)))))

(define (modep p l)
 ;; belongs in QobiScheme
 (first
  (first				;arbitrary
   (minimal-members (lambda (x y) (> (length x) (length y)))
		    (lambda (x y) (>= (length x) (length y)))
		    (equivalence-classesp p l)))))

(define (multiset-intersectionq s1 s2)
 ;; belongs in QobiScheme
 (let ((p1 (equivalence-classes s1))
       (p2 (equivalence-classes s2)))
  (reduce
   append
   (removeq
    #f
    (map (lambda (c1)
	  (let ((c2 (find-if (lambda (c2) (eq? (first c1) (first c2))) p2)))
	   (if c2 (if (< (length c1) (length c2)) c1 c2) #f)))
	 p1))
   '())))

(define (multiset-intersectionv s1 s2)
 ;; belongs in QobiScheme
 (let ((p1 (equivalence-classes s1))
       (p2 (equivalence-classes s2)))
  (reduce
   append
   (removeq
    #f
    (map (lambda (c1)
	  (let ((c2 (find-if (lambda (c2) (eqv? (first c1) (first c2))) p2)))
	   (if c2 (if (< (length c1) (length c2)) c1 c2) #f)))
	 p1))
   '())))

(define (multiset-intersection s1 s2)
 ;; belongs in QobiScheme
 (let ((p1 (equivalence-classes s1))
       (p2 (equivalence-classes s2)))
  (reduce
   append
   (removeq
    #f
    (map (lambda (c1)
	  (let ((c2 (find-if (lambda (c2) (equal? (first c1) (first c2))) p2)))
	   (if c2 (if (< (length c1) (length c2)) c1 c2) #f)))
	 p1))
   '())))

(define (multiset-intersectionp p? s1 s2)
 ;; belongs in QobiScheme
 (let ((p1 (equivalence-classes s1))
       (p2 (equivalence-classes s2)))
  (reduce
   append
   (removeq
    #f
    (map (lambda (c1)
	  (let ((c2 (find-if (lambda (c2) (p? (first c1) (first c2))) p2)))
	   (if c2 (if (< (length c1) (length c2)) c1 c2) #f)))
	 p1))
   '())))

(define (remove-oneq x l)
 ;; belongs in QobiScheme
 ;; needs work: To eliminate REVERSE.
 (let loop ((l l) (c '()))
  (cond ((null? l) (reverse c))
	((eq? x (first l)) (append (reverse c) (rest l)))
	(else (loop (rest l) (cons (first l) c))))))

(define (remove-onev x l)
 ;; belongs in QobiScheme
 ;; needs work: To eliminate REVERSE.
 (let loop ((l l) (c '()))
  (cond ((null? l) (reverse c))
	((eqv? x (first l)) (append (reverse c) (rest l)))
	(else (loop (rest l) (cons (first l) c))))))

(define (remove-one x l)
 ;; belongs in QobiScheme
 ;; needs work: To eliminate REVERSE.
 (let loop ((l l) (c '()))
  (cond ((null? l) (reverse c))
	((equal? x (first l)) (append (reverse c) (rest l)))
	(else (loop (rest l) (cons (first l) c))))))

(define (remove-onep p? x l)
 ;; belongs in QobiScheme
 ;; needs work: To eliminate REVERSE.
 (let loop ((l l) (c '()))
  (cond ((null? l) (reverse c))
	((p? x (first l)) (append (reverse c) (rest l)))
	(else (loop (rest l) (cons (first l) c))))))

(define (multiset-differenceq s1 s2)
 ;; belongs in QobiScheme
 (if (null? s2)
     s1
     (multiset-differenceq (remove-oneq (first s2) s1) (rest s2))))

(define (multiset-differencev s1 s2)
 ;; belongs in QobiScheme
 (if (null? s2)
     s1
     (multiset-differencev (remove-onev (first s2) s1) (rest s2))))

(define (multiset-difference s1 s2)
 ;; belongs in QobiScheme
 (if (null? s2)
     s1
     (multiset-difference (remove-one (first s2) s1) (rest s2))))

(define (multiset-differencep p? s1 s2)
 ;; belongs in QobiScheme
 (if (null? s2)
     s1
     (multiset-differencep p? (remove-onep p? (first s2) s1) (rest s2))))

(define (internalize-term t)
 (cond
  ((pair? t) (map internalize-term t))
  ((ama:variable? t) (string->symbol (format "?~s" (ama:variable-count t))))
  (else t)))

(define (variables-in t)
 (cond ((pair? t) (reduce unionq (map variables-in t) '()))
       ((ama:variable? t) (list t))
       (else '())))

(define (constants-in t)
 (cond ((pair? t) (reduce unionq (map constants-in t) '()))
       ((ama:variable? t) '())
       (else (list t))))

(define (externalize-term t)
 (let ((variables (variables-in t)))
  (let loop ((t t))
   (cond
    ((pair? t) (map loop t))
    ((ama:variable? t) (string->symbol (format "?~s" (positionq t variables))))
    (else t)))))

(define (ama:apply-substitution s t)
 (cond
  ((pair? t)
   (cons (first t) (map (lambda (t) (ama:apply-substitution s t)) (rest t))))
  ((assq t s) (cdr (assq t s)))
  (else t)))

(define (lift t)
 (let ((constants (constants-in t)))
  (ama:apply-substitution
   (map (lambda (constant) (cons constant (create-variable constant constant)))
	constants)
   t)))

(define (augment-substitution b s)
 (and (not (eq? s #f))
      (let ((b1 (assoc (car b) s)))
       (if b1 (if (equal? (cdr b1) (cdr b)) s #f) (cons b s)))))

(define (merge-substitutions s1 s2)
 (if (null? s1)
     s2
     (merge-substitutions (rest s1) (augment-substitution (first s1) s2))))

(define (is-variable? t)
 (or (ama:variable? t)
     (memq t '(w x y z))
     (and (symbol? t)
	  (>= (string-length (symbol->string t)) 1)
	  (char=? (string-ref (symbol->string t) 0) #\?))))

(define (term-semiunify t1 t2 s)
 ;; Assumes that the variables in T1 and T2 are disjoint so there is no need
 ;; for an occurs check.
 (let outer ((t1 t1) (t2 t2) (s s))
  (cond ((list? t1)
	 (cond ((list? t2)
		;; Assumes lists are not empty and are headed by symbols.
		(and (eq? (first t1) (first t2))
		     (= (length t1) (length t2))
		     (let inner ((ts1 (rest t1)) (ts2 (rest t2)) (s s))
		      (if (null? ts1)
			  s
			  (inner (rest ts1)
				 (rest ts2)
				 (outer (first ts1) (first ts2) s))))))
	       ((is-variable? t2) (augment-substitution (cons t2 t1) s))
	       ;; Assumes non-lists are variables or symbols.
	       (else #f)))
	((is-variable? t1)
	 (cond ((list? t2) #f)
	       ((is-variable? t2) (augment-substitution (cons t2 t1) s))
	       ;; Assumes non-lists are variables or symbols.
	       (else #f)))
	;; Assumes non-lists are variables or symbols.
	(else (cond ((list? t2) #f)
		    ((is-variable? t2) (augment-substitution (cons t2 t1) s))
		    ;; Assumes non-lists are variables or symbols.
		    (else (if (eq? t1 t2) s #f)))))))

(define (term-less-general-or-equivalent? t1 t2)
 (not (eq? (term-semiunify t1 t2 '()) #f)))

(define (term-less-general? t1 t2)
 (and (term-less-general-or-equivalent? t1 t2)
      (not (term-less-general-or-equivalent? t2 t1))))

(define (create-variable t1 t2)
 (let ((variable (find-if (lambda (variable)
			   (and (equal? (ama:variable-t1 variable) t1)
				(equal? (ama:variable-t2 variable) t2)))
			  *variables*)))
  (unless variable
   (set! variable (make-ama:variable t1 t2 *variable-counter*))
   (set! *variables* (cons variable *variables*))
   (set! *variable-counter* (+ *variable-counter* 1)))
  variable))

(define (term-lgg? t1 t2)
 ;; Assumes that the variables in T1 and T2 are disjoint so there is no need
 ;; for an occurs check.
 (let loop ((t1 t1) (t2 t2))
  (or (not (list? t1))
      (not (list? t2))
      ;; Assumes lists are not empty and are headed by symbols.
      (and (eq? (first t1) (first t2))
	   (= (length t1) (length t2))
	   (every loop (rest t1) (rest t2))))))

(define (term-lgg t1 t2)
 ;; Assumes that the variables in T1 and T2 are disjoint so there is no need
 ;; for an occurs check.
 (let loop ((t1 t1) (t2 t2))
  (if (list? t1)
      (cond ((list? t2)
	     ;; Assumes lists are not empty and are headed by symbols.
	     (unless (and (eq? (first t1) (first t2))
			  (= (length t1) (length t2)))
	      (fail))
	     (cons (first t1) (map loop (rest t1) (rest t2))))
	    (else (create-variable t1 t2)))
      (if (list? t2)
	  (create-variable t1 t2)
	  ;; Assumes non-lists are symbols.
	  (if (eq? t1 t2) t1 (create-variable t1 t2))))))

(define (terms-lgg ts)
 (when (null? ts) (fuck-up))
 ;; The #T below is a place holder that should never be used.
 (reduce term-lgg ts #t))

(define (conjunction-less-general-or-equivalent? a1 a2)
 (if *lifted?*
     (possibly?
      (let loop ((ts2 (rest a2)) (s '()))
       (if (null? ts2)
	   #t
	   (let* ((t1 (a-member-of (rest a1)))
		  (s (term-semiunify t1 (first ts2) s)))
	    (when (eq? s #f) (fail))
	    (loop (rest ts2) s)))))
     (subset? a2 a1)))

(define (conjunction-less-general? a1 a2)
 (and (conjunction-less-general-or-equivalent? a1 a2)
      (not (conjunction-less-general-or-equivalent? a2 a1))))

(define (conjunction-lgg a1 a2)
 (if *lifted?*
     (cons
      'and
      (minimal-values
       term-less-general?
       term-less-general-or-equivalent?
       (lambda () (term-lgg (a-member-of (rest a1)) (a-member-of (rest a2))))))
     (cons 'and (intersection (rest a1) (rest a2)))))

(define (conjunctions-lgg as)
 (when (null? as) (fuck-up))
 ;; The #T below is a place holder that should never be used.
 (reduce conjunction-lgg as #t))

(define (merge-adjacent ma)
 (cons 'begin
       (let loop ((as (rest ma)))
	(cond ((or (null? as) (null? (rest as))) as)
	      ((set-equal? (first as) (second as)) (loop (rest as)))
	      (else (cons (first as) (loop (rest as))))))))

(define (ama:simplify ama) (cons 'and (remove '(begin (and)) (rest ama))))

(define (a-list-partition-of l)
 (when (null? l) (fail))
 (let loop ((l (rest l)) (p (list (list (first l)))))
  (if (null? l)
      (reverse (map reverse p))
      (loop (rest l)
	    (either (cons (list (first l)) p)
		    (cons (cons (first l) (first p)) (rest p)))))))

(define (a-list-partition-of-size k l)
 (when (null? l) (fail))
 (let loop ((l (rest l)) (p (list (list (first l)))))
  (when (or (> (length p) k) (< (+ (length l) (length p)) k)) (fail))
  (if (null? l)
      (reverse (map reverse p))
      (loop (rest l)
	    (either (cons (list (first l)) p)
		    (cons (cons (first l) (first p)) (rest p)))))))

(define (an-interdigitation-of l1 l2)
 (when (or (null? l1) (null? l2)) (fail))
 (let loop ((l1 l1) (l2 l2))
  (if (null? (rest l1))
      (if (null? (rest l2))
	  (list (list (first l1) (first l2)))
	  (cons (list (first l1) (first l2)) (loop l1 (rest l2))))
      (if (null? (rest l2))
	  (cons (list (first l1) (first l2)) (loop (rest l1) l2))
	  (either
	   (cons (list (first l1) (first l2)) (loop l1 (rest l2)))
	   (cons (list (first l1) (first l2)) (loop (rest l1) l2))
	   (cons (list (first l1) (first l2)) (loop (rest l1) (rest l2))))))))

(define (an-interdigitation-of-size k l1 l2)
 (when (or (null? l1) (null? l2)) (fail))
 (let loop ((k k) (l1 l1) (l2 l2))
  (if (= k 1)
      (list (list l1 l2))
      (let ((n1 (length l1)) (n2 (length l2)))
       (either
	(list (list l1 l2))
	(let ((i1 (an-integer-between 1 (- n1 1)))
	      (i2 (an-integer-between 1 (- n2 1))))
	 (cons (list (sublist l1 0 i1) (sublist l2 0 i2))
	       (loop (- k 1) (sublist l1 i1 n1) (sublist l2 i2 n2))))
	(let ((i1 (an-integer-between 1 n1))
	      (i2 (an-integer-between 1 (- n2 1))))
	 (cons (list (sublist l1 0 i1) (sublist l2 0 i2))
	       (loop (- k 1) (sublist l1 (- i1 1) n1) (sublist l2 i2 n2))))
	(let ((i1 (an-integer-between 1 (- n1 1)))
	      (i2 (an-integer-between 1 n2)))
	 (cons
	  (list (sublist l1 0 i1) (sublist l2 0 i2))
	  (loop (- k 1) (sublist l1 i1 n1) (sublist l2 (- i2 1) n2)))))))))

(define (irredundant-interdigitation? pairs)
 (or
  (null? (rest pairs))
  (every
   (lambda (pair1 pair2)
    (and (or (and (null? (rest (first pair1))) (null? (rest (first pair2))))
	     (not (equal? (last (first pair1)) (first (first pair2)))))
	 (or (and (null? (rest (second pair1))) (null? (rest (second pair2))))
	     (not (equal? (last (second pair1)) (first (second pair2)))))))
   (but-last pairs)
   (rest pairs))))

(define (an-irredundant-interdigitation-of-size k l1 l2)
 (define (loop k l1 l2 p1? p2?)
  (cond
   ((= k 1) (list (list l1 l2)))
   ((= k 2)
    (let ((n1 (length l1)) (n2 (length l2)))
     (either
      (let ((i1 (if p1? 1 (an-integer-between 1 (- n1 1))))
	    (i2 (if p2? 1 (an-integer-between 1 (- n2 1)))))
       (cons (list (sublist l1 0 i1) (sublist l2 0 i2))
	     (loop (- k 1) (sublist l1 i1 n1) (sublist l2 i2 n2) #f #f)))
      (begin (unless (= n1 1) (fail))
	     (let ((i2 (if p2? 1 (an-integer-between 1 (- n2 1)))))
	      (cons (list (list (first l1)) (sublist l2 0 i2))
		    (loop1 (- k 1) l1 (sublist l2 i2 n2)))))
      (begin (unless (= n2 1) (fail))
	     (let ((i1 (if p1? 1 (an-integer-between 1 (- n1 1)))))
	      (cons (list (sublist l1 0 i1) (list (first l2)))
		    (loop2 (- k 1) (sublist l1 i1 n1) l2)))))))
   (else (let ((n1 (length l1)) (n2 (length l2)))
	  (either
	   (let ((i1 (if p1? 1 (an-integer-between 1 (- n1 1))))
		 (i2 (if p2? 1 (an-integer-between 1 (- n2 1)))))
	    (cons (list (sublist l1 0 i1) (sublist l2 0 i2))
		  (loop (- k 1) (sublist l1 i1 n1) (sublist l2 i2 n2) #f #f)))
	   (let ((i2 (if p2? 1 (an-integer-between 1 (- n2 1)))))
	    (cons (list (list (first l1)) (sublist l2 0 i2))
		  (if (= n1 1)
		      (loop1 (- k 1) l1 (sublist l2 i2 n2))
		      (loop (- k 1) l1 (sublist l2 i2 n2) #t #f))))
	   (let ((i1 (if p1? 1 (an-integer-between 1 (- n1 1)))))
	    (cons (list (sublist l1 0 i1) (list (first l2)))
		  (if (= n2 1)
		      (loop2 (- k 1) (sublist l1 i1 n1) l2)
		      (loop (- k 1) (sublist l1 i1 n1) l2 #f #t)))))))))
 (define (loop1 k l1 l2)
  (if (= k 1)
      (list (list l1 l2))
      (let* ((n2 (length l2)) (i2 (an-integer-between 1 (- n2 1))))
       (cons (list (list (first l1)) (sublist l2 0 i2))
	     (loop1 (- k 1) l1 (sublist l2 i2 n2))))))
 (define (loop2 k l1 l2)
  (if (= k 1)
      (list (list l1 l2))
      (let* ((n1 (length l1)) (i1 (an-integer-between 1 (- n1 1))))
       (cons (list (sublist l1 0 i1) (list (first l2)))
	     (loop2 (- k 1) (sublist l1 i1 n1) l2)))))
 (when (or (null? l1) (null? l2)) (fail))
 (loop (an-integer-between 1 k) l1 l2 #f #f))

(define (ma-less-general-or-equivalent? ma1 ma2)
 (if *lifted?*
     (possibly?
      (every
       (lambda (pair)
	(conjunction-less-general-or-equivalent? (first pair) (second pair)))
       (an-interdigitation-of (rest ma1) (rest ma2))))
     (let* ((n1 (length (rest ma1)))
	    (n2 (length (rest ma2)))
	    (v1 (list->vector (rest ma1)))
	    (v2 (list->vector (rest ma2)))
	    (m (make-matrix n2 n1 0))
	    ;; FLAG? is used for an early pruning test
	    (flag? #f))
      (let loop ((col 0))		; compute first row of matrix
       (when (and (< col n1)
		  (conjunction-less-general-or-equivalent?
		   (vector-ref v1 col) (vector-ref v2 0)))
	(matrix-set! m 0 col 1)
	(loop (+ col 1))))
      (cond
       ((or (= (matrix-ref m 0 0) 0)	; early exit test
	    (not (conjunction-less-general-or-equivalent?
		  (vector-ref v1 (- n1 1)) (vector-ref v2 (- n2 1)))))
	#f)
       (else
	(let loop ((row 1))		; compute the first column
	 (when (and (< row n2)
		    (conjunction-less-general-or-equivalent?
		     (vector-ref v1 0) (vector-ref v2 row)))
	  (matrix-set! m row 0 1)
	  (loop (+ row 1))))
	(let loop-row ((row 1))		; search row-wise for a path
	 (cond ((< row n2)
		(set! flag? (= (matrix-ref m row 0) 1))
		(let loop-col ((col 1))
		 (cond ((< col n1)
			(when (and (or (= (matrix-ref m (- row 1) (- col 1)) 1)
				       (= (matrix-ref m (- row 1) col) 1)
				       (= (matrix-ref m row (- col 1)) 1))
				   (conjunction-less-general-or-equivalent?
				    (vector-ref v1 col)
				    (vector-ref v2 row)))
			 (matrix-set! m row col 1)
			 (set! flag? #t))
			(loop-col (+ col 1)))
		       (else (if flag? (loop-row (+ row 1)) (loop-row n2))))))
	       (else (= (matrix-ref m (- n2 1) (- n1 1)) 1)))))))))

(define (ma-less-general? ma1 ma2)
 (and (ma-less-general-or-equivalent? ma1 ma2)
      (not (ma-less-general-or-equivalent? ma2 ma1))))

(define (k-l-ama-generalization k l ama)
 (ama:simplify
  (cons
   'and
   (first-n
    l
    (minimal-values
     ma-less-general?
     ma-less-general-or-equivalent?
     (lambda ()
      (merge-adjacent
       (let ((ma (a-member-of (rest ama))))
	(if (< (length (rest ma)) k)
	    ma
	    (cons
	     'begin
	     ;; The #T below is a place holder that should never be used.
	     (map (lambda (pss) (reduce intersection pss #t))
		  (a-list-partition-of-size k (rest ma)))))))))))))

(define (ama-lgg ama1 ama2)
 (ama:simplify
  (cons
   'and
   (minimal-values
    ma-less-general?
    ma-less-general-or-equivalent?
    (lambda ()
     (merge-adjacent
      (cons
       'begin
       (map (lambda (pair) (conjunction-lgg (first pair) (second pair)))
	    (an-interdigitation-of (rest (a-member-of (rest ama1)))
				   (rest (a-member-of (rest ama2))))))))))))

(define (original-k-l-ama-lgg k l ama1 ama2)
 (ama:simplify
  (cons
   'and
   (first-n
    l
    (minimal-values
     ma-less-general?
     ma-less-general-or-equivalent?
     (lambda ()
      (merge-adjacent
       (cons
	'begin
	(map
	 (lambda (pair)
	  ;; The #Ts below are place holders that should never be used.
	  (conjunction-lgg (reduce intersection (first pair) #t)
			   (reduce intersection (second pair) #t)))
	 (an-irredundant-interdigitation-of-size
	  k
	  (rest (a-member-of (rest ama1)))
	  (rest (a-member-of (rest ama2)))))))))))))

;;; afern-revision

(define (k-l-ama-lgg k l ama1 ama2)
  (let* ((subsumer-indices (ama-subsumers ama1 ama2))
         (subsumers (append (map (lambda (i) (list-ref (rest ama1) i)) 
                                 (first subsumer-indices))
                            (map (lambda (i) (list-ref (rest ama2) i)) 
                                 (second subsumer-indices))))
         (mas1 (map (lambda (i) (list-ref (rest ama1) i))
                    (set-difference (enumerate (length (rest ama1)))
                                    (first subsumer-indices))))
         (mas2 (map (lambda (i) (list-ref (rest ama2) i))
                    (set-difference (enumerate (length (rest ama2)))
                                    (second subsumer-indices)))))
;    (pp (list (length mas1) (length mas2)))
    (ama:simplify
     (cons
      'and
      (first-n
       l 
       (minimal-values
        ma-less-general?
        ma-less-general-or-equivalent?
        (lambda ()
          (if (a-boolean)
              (a-member-of subsumers)
              (merge-adjacent
               (cons
                'begin
                (map
                 (lambda (pair)
	  ;; The #Ts below are place holders that should never be used.
                   (conjunction-lgg (reduce intersection (first pair) #t)
                                    (reduce intersection (second pair) #t)))
                 (an-irredundant-interdigitation-of-size
                  k
                  (rest (a-member-of mas1))
                  (rest (a-member-of mas2))))))))))))))

;; afern-revision
  
(define (ama-subsumers ama1 ama2)
  (let* ((mas1 (cdr ama1))
         (mas2 (cdr ama2))
         (s1 '())
         (s2 '()))
   (for-each
     (lambda (i1) 
       (for-each 
        (lambda (i2)
          (cond ((ma-less-general-or-equivalent? 
                  (list-ref mas1 i1) (list-ref mas2 i2))
                 (set! s2 (append (list i2) s2)))
                ((ma-less-general-or-equivalent? 
                  (list-ref mas2 i2) (list-ref mas1 i1))
                 (set! s1 (append (list i1) s1)))))
       (enumerate (length mas2))))
    (enumerate (length mas1)))
  (list (remove-duplicates s1) (remove-duplicates s2))))

(define (amas-lgg amas)
 ;; The #T below is a place holder that should never be used.
 (reduce ama-lgg amas #t))

(define (k-l-amas-lgg k l amas)
 (reduce
  (lambda (ama1 ama2)
   ;; debugging
   (display (format "LGG: ~s ~s~%" (length (rest ama1)) (length (rest ama2))))
   (k-l-ama-lgg k l ama1 ama2))
  amas
  ;; The #T below is a place holder that should never be used.
  #t))

(define (read-models)
 (call-with-input-file "/home/qobi/kr2002/models-all.text"
  (lambda (port)
   (let loop ((ms '()))
    (let ((m (read port)))
     (if (eof-object? m)
	 (reverse ms)
	 (loop
	  (cons
	   (map (lambda (event-occurrence)
		 (list (first event-occurrence)
		       (map (lambda (i)
			     (make-extended-spanning-interval
			      (vector-ref i 1)
			      (vector-ref i 2)
			      (vector-ref i 3)
			      (vector-ref i 4)
			      (vector-ref i 5)
			      (vector-ref i 6)
			      (vector-ref i 7)
			      (vector-ref i 8)
			      (vector-ref i 9)
			      (vector-ref i 10)))
			    (second event-occurrence))))
		m)
	   ms))))))))

(define (split-occurrences ms)
 (map (lambda (m)
       (reduce append
	       (map (lambda (event-occurrence)
		     (map (lambda (extended-spanning-interval)
			   (list (first event-occurrence)
				 (list extended-spanning-interval)))
			  (second event-occurrence)))
		    m)
	       '()))
      ms))

(define (primitive-event-occurrence? event-occurrence)
 (memq (first (first event-occurrence))
       '(supports? contacts? attached? supported?)))

(define (only-primitive-events ms)
 (map (lambda (m) (remove-if-not primitive-event-occurrence? m)) ms))

(define (split-models ms)
 (reduce append
	 (map (lambda (m)
	       (let ((primitive-event-occurrences
		      (remove-if-not primitive-event-occurrence? m)))
		(map (lambda (compound-event-occurrence)
		      (cons compound-event-occurrence
			    primitive-event-occurrences))
		     (remove-if primitive-event-occurrence? m))))
	      ms)
	 '()))

(define (models-for event-type)
 (remove-if-not (lambda (m)
		 ;; This assumes that the first event occurrence of a model is
		 ;; a compound event occurrence and the only compound event
		 ;; occurrence.
		 (some (lambda (event-occurrence)
			(eq? (first (first event-occurrence)) event-type))
		       m))
		(split-models (split-occurrences (read-models)))))

(define (string->extended-spanning-interval string)
 (when (or (memq #\i (string->list string))
	   (memq #\I (string->list string))
	   (memq #\{ (string->list string))
	   (memq #\} (string->list string)))
  (panic "Can't (yet) parse this spanning interval"))
 (if (memq #\: (string->list string))
     (let ((colon (positionq #\: (string->list string)))
	   (n (string-length string)))
      (first
       (extended-spanning-intervals
	(string->number (substring string 2 colon))
	(string->number (substring string (+ colon 1) (- n 2)))
	(string->number (substring string 2 colon))
	(string->number (substring string (+ colon 1) (- n 2)))
	(char=? (string-ref string 0) #\[)
	(char=? (string-ref string (- n 1)) #\])
	(char=? (string-ref string 1) #\[)
	(char=? (string-ref string (- n 2)) #\])
	(char=? (string-ref string 1) #\[)
	(char=? (string-ref string (- n 2)) #\]))))
     (let* ((n (string-length string))
	    (comma
	     (+ (positionq #\, (string->list string))
		1
		(positionq
		 #\,
		 (string->list
		  (substring string
			     (+ (positionq #\, (string->list string)) 1)
			     n)))))
	    (start (substring string 1 comma))
	    (end (substring string (+ comma 1) (- n 1)))
	    (start-comma (positionq #\, (string->list start)))
	    (end-comma (positionq #\, (string->list end)))
	    (start-n (string-length start))
	    (end-n (string-length end)))
      (make-extended-spanning-interval
       (string->number (substring start 1 start-comma))
       (string->number (substring start (+ start-comma 1) (- start-n 1)))
       (string->number (substring end 1 end-comma))
       (string->number (substring end (+ end-comma 1) (- end-n 1)))
       (char=? (string-ref string 0) #\[)
       (char=? (string-ref string (- n 1)) #\])
       (char=? (string-ref start 0) #\[)
       (char=? (string-ref start (- start-n 1)) #\])
       (char=? (string-ref end 0) #\[)
       (char=? (string-ref end (- end-n 1)) #\])))))

(define (string->event-occurrence string)
 (list
  (let ((port (open-input-string
	       (substring string 0 (positionq #\@ (string->list string))))))
   (read port))
  (let ((fields
	 (fields
	  (substring string
		     (+ (positionq #\@ (string->list string)) 2)
		     (- (string-length string) 1)))))
   (map
    string->extended-spanning-interval
    (append
     (map (lambda (field) (substring field 0 (- (string-length field) 1)))
	  (but-last fields))
     (list (last fields)))))))

(define (read-event-file pathname)
 (map string->event-occurrence
      (remove-if (lambda (string) (zero? (string-length string)))
		 (read-file pathname))))

(define (event-pathname corpus subject event-type condition index)
 (format
  "~a/~a-~a-~a~a-events.text"
  *howard-directory* event-type condition subject index))

;;; afern-revision

(define (read-models-of-type-and-condition corpus subject 
                                           event-type condition)
  (let* ((ms (map (lambda (index)
                    (read-event-file
                     (event-pathname corpus subject 
                                     event-type condition index)))
                  *indices*))
         (constants (reduce append
                            (map constants-in-model ms) #f)))
    (map 
     (lambda (m)
       (map 
        (lambda (e)
          (cons (if (or (equal? 'contacts? (caar e))
                        (equal? 'attached? (caar e)))
                    (let ((a (second (car e)))
                          (b (third (car e))))
                      (if (before-in-list a b constants)
                          (list (caar e) a b)
                          (list (caar e) b a)))
                    (car e))
                (cdr e)))
        m))
     ms)))

(define (before-in-list a b l)
  (cond ((null? l)
	 #f)
	((equal? a (car l))
	 #t)
	((equal? b (car l))
	 #f)
	(else (before-in-list a b (cdr l)))))

(define (prev-read-models-of-type-and-condition corpus subject event-type condition)
 (map (lambda (index)
       (read-event-file
	(event-pathname corpus subject event-type condition index)))
      *indices*))

(define (read-models-of-type corpus subject event-type)
 (reduce append
	 (map (lambda (condition)
	       (read-models-of-type-and-condition
		corpus subject event-type condition))
	      *conditions*)
	 '()))

(define (read-all-models corpus subject)
 (reduce append
	 (map (lambda (event-type)
	       (read-models-of-type corpus subject event-type))
	      *event-types*)
	 '()))

(define (term-map-constants t mapping)
 (let loop ((t t))
  (if (list? t)
      (cons (first t) (map loop (rest t)))
      (let ((b (assq t mapping))) (if b (cdr b) t)))))

(define (model-map-constants m mapping)
 (map (lambda (event-occurrence)
       (list (term-map-constants (first event-occurrence) mapping)
	     (second event-occurrence)))
      m))

(define (constants-in-term t)
 (let loop ((t t))
  (if (list? t)
      (reduce unionq (map loop (rest t)) '())
      (if (ama:variable? t) '() (list t)))))

(define (constants-in-model m)
 (reduce
  unionq
  (map (lambda (event-occurrence) (constants-in-term (first event-occurrence)))
       m)
  '()))

(define (term-canonical-mapping t)
 ;; This assumes that there are no function symbols or variables as
 ;; arguments to the term.
 ;; It also assumes that there is no repetition of constants as argument to
 ;; that term.
 ;; It also assumes that there are at most four arguments to that term.
 (map cons (append (rest t) (set-differenceq '(a b c d) (rest t))) '(a b c d)))

(define (model-canonical-mapping m)
 ;; This assumes that the first event occurrence of a model is a compound
 ;; event occurrence and the only compound event occurrence.
 (term-canonical-mapping (first (first m))))

(define (canonize-terms ts)
 ;; This assumes that the first term is a compound event and the only compound
 ;; event.
 (rest (term-map-constants (cons 'f ts) (term-canonical-mapping (first ts)))))

(define (canonize-model m) (model-map-constants m (model-canonical-mapping m)))

(define (compose-mapping mapping1 mapping2)
 (map (lambda (pair1)
       (let ((pair2 (assq (cdr pair1) mapping2)))
	(unless pair2 (panic "Can't compose mappings"))
	(cons (car pair1) (cdr pair2))))
      mapping1))

(define (composable-mapping? mapping1 mapping2)
 (every (lambda (pair1) (assq (cdr pair1) mapping2)) mapping1))

(define (internalize-model m)
 ;; This assumes that the first event occurrence is a compound event and the
 ;; only compound event.
 (let ((i (extended-spanning-interval-i (first (second (first m)))))
       (l (extended-spanning-interval-l (first (second (first m))))))
  (remove-if
   (lambda (event-occurrence) (null? (second event-occurrence)))
   (map (lambda (event-occurrence)
	 (list (first event-occurrence)
	       (reduce append
		       (map (lambda (k)
			     (extended-spanning-intervals
			      (max (extended-spanning-interval-i k) i)
			      (extended-spanning-interval-j k)
			      (extended-spanning-interval-k k)
			      (min (extended-spanning-interval-l k) l)
			      (extended-spanning-interval-alpha? k)
			      (extended-spanning-interval-beta? k)
			      #t
			      #t
			      #t
			      #t))
			    (second event-occurrence))
		       '())))
	(remove-if-not (lambda (event-occurrence)
			(let ((k (first (second event-occurrence))))
			 (and (<= (extended-spanning-interval-i k) l)
			      (<= i (extended-spanning-interval-l k)))))
		       m)))))

(define (model-points m)
 (sort (reduce union
	       (map (lambda (event-occurrence)
		     (let ((i (first (second event-occurrence))))
		      (remove-duplicates
		       (list (extended-spanning-interval-i i)
			     (extended-spanning-interval-j i)
			     (extended-spanning-interval-k i)
			     (extended-spanning-interval-l i)))))
		    (remove-if-not primitive-event-occurrence? m))
	       '())
       <
       identity))

(define (model->ama m)
 (let ((is (model-points m)))
  (list 'and
	(cons 'begin
	      (map (lambda (i j)
		    (cons
		     'and
		     (map first
			  (remove-if-not
			   (lambda (event-occurrence)
			    (let ((k (first (second event-occurrence))))
			     (and (<= (extended-spanning-interval-i k) i)
				  (<= j (extended-spanning-interval-l k)))))
			   (remove-if-not primitive-event-occurrence? m)))))
		   (but-last is)
		   (rest is))))))

(define (all-ground-propositions m)
 ;; This procedure is specific to Leonard. It also won't work with multiple
 ;; moving objects.
 (let ((constants (constants-in-model m)))
  (define (a-constant) (a-member-of constants))
  (all-values
   (either `(supported? ,(a-constant))
	   (let ((x (a-constant)) (y (a-constant)))
	    (when (eq? x y) (fail))
	    `(supports? ,x ,y))
	   (let ((x (a-constant)) (y (a-constant)))
	    (when (memq x (memq y constants)) (fail))
	    `(contacts? ,x ,y))
	   (let ((x (a-constant)) (y (a-constant)))
	    (when (memq x (memq y constants)) (fail))
	    `(attached? ,x ,y))))))

(define (model->ama-n m)
 (let ((is (model-points m)) (ts (all-ground-propositions m)))
  (list
   'and
   (cons 'begin
	 (map (lambda (i j)
	       (let ((positive-literals
		      (map first
			   (remove-if-not
			    (lambda (event-occurrence)
			     (let ((k (first (second event-occurrence))))
			      (and (<= (extended-spanning-interval-i k) i)
				   (<= j (extended-spanning-interval-l k)))))
			    (remove-if-not primitive-event-occurrence? m)))))
		(cons
		 'and
		 (append
		  positive-literals
		  (map (lambda (t) `(not (diamond (d) ,t)))
		       (set-difference ts positive-literals))))))
	      (but-last is)
	      (rest is))))))

;;; afern-revision
(define (prev-partial-model->ama-n m)
 (let ((is (model-points m)) (ts (map first m)))
  (list
   'and
   (cons 'begin
	 (map (lambda (i j)
	       (let ((positive-literals
		      (map first
			   (remove-if-not
			    (lambda (event-occurrence)
			     (let ((k (first (second event-occurrence))))
			      (and (<= (extended-spanning-interval-i k) i)
				   (<= j (extended-spanning-interval-l k)))))
			    (remove-if-not primitive-event-occurrence? m)))))
		(cons
		 'and
		 (append
		  positive-literals
		  (map (lambda (t) `(not (diamond (d) ,t)))
		       (set-difference ts positive-literals))))))
	      (but-last is)
	      (rest is))))))

(define (partial-model->ama-n m)
 (let ((is (model-points m)) (ts (map first m)))
  (list
   'and
   (cons 'begin
	 (map (lambda (i j)
	       (let ((positive-literals
		      (map first
			   (remove-if-not
			    (lambda (event-occurrence)
			     (let ((k (first (second event-occurrence))))
			      (and (<= (extended-spanning-interval-i k) i)
				   (<= j (extended-spanning-interval-l k)))))
			    (remove-if-not primitive-event-occurrence? m)))))
                 (cons
		 'and
		 (append
		  positive-literals
		  (map (lambda (t) `(not (diamond (d) ,t)))
		       (remove-duplicates 
                        (set-difference ts positive-literals)))))))
	      (but-last is)
	      (rest is))))))

(define (partial-model->ama-pre-post-n m)
  (let ((ma (second (partial-model->ama-n m))))
    (list 'and
          (append (list (first ma) (second ma))
                  (let loop ((l (rest (rest ma))))
                    (if (= (length l) 1)
                        l
                        (cons (remove-if (lambda (e) 
                                           (if (pair? e)
                                               (equal? (first e) 'not)
                                               #f))
                                         (first l))
                              (loop (rest l)))))))))    

(define (fast-occurrences-in-model e m)
 (define (make-super-irredundant occurrences)
  (reduce
   append
   (map (lambda (occurrences)
	 (map (lambda (extended-spanning-interval)
	       (list (first (first occurrences)) extended-spanning-interval))
	      (make-irredundant (map second occurrences))))
	(equivalence-classesp
	 (lambda (occurrence1 occurrence2)
	  (set-equal? (first occurrence1) (first occurrence2)))
	 occurrences))
   '()))
 (define (merge r e1 e2)
  (let ((occurrences1 (fast-occurrences-in-model e1 m))
	(occurrences2 (fast-occurrences-in-model e2 m)))
   (make-super-irredundant
    (reduce
     append
     (map
      (lambda (occurrence1)
       (reduce
	append
	(map
	 (lambda (occurrence2)
	  (let ((s (merge-substitutions
		    (first occurrence1) (first occurrence2))))
	   (if (eq? s #f)
	       '()
	       (map (lambda (extended-spanning-interval)
		     (list s extended-spanning-interval))
		    (span-allen (list (second occurrence1))
				r
				(list (second occurrence2)))))))
	 occurrences2)
	'()))
      occurrences1)
     '()))))
 (case (first e)
  ((not) (panic "NOT unimplemented"))
  ((or) (make-super-irredundant
	 (reduce
	  append
	  (map (lambda (e) (fast-occurrences-in-model e m)) (rest e))
	  '())))
  ((every) (panic "EVERY unimplemented"))
  ((some) (panic "SOME unimplemented"))
  ((and)
   (when (and (not (null? (rest e)))
	      (list? (second e))
	      (subsetq? (second e) '(= < > m mi o oi s si f fi d di)))
    (panic "AND with subscript R unimplemented"))
   (if (null? (rest e))
       (map (lambda (extended-spanning-interval)
	     (list '() extended-spanning-interval))
	    (all-intervals))
       (merge '= (second e) `(and ,@(rest (rest e))))))
  ((begin) (if (null? (rest (rest e)))
	       (fast-occurrences-in-model (second e) m)
	       (merge 'm (second e) `(begin ,@(rest (rest e))))))
  ((diamond) (panic "DIAMOND unimplemented"))
  ((equal?) (panic "EQUAL? unimplemented"))
  ((supported?)
   (reduce
    append
    (map (lambda (event-occurrence)
	  (let ((s (term-semiunify (first event-occurrence) e '())))
	   (if (eq? s #f)
	       '()
	       (map (lambda (extended-spanning-interval)
		     (list s extended-spanning-interval))
		    (second event-occurrence)))))
	 m)
    '()))
  ((supports?)
   (reduce
    append
    (map (lambda (event-occurrence)
	  (let ((s (term-semiunify (first event-occurrence) e '())))
	   (if (eq? s #f)
	       '()
	       (map (lambda (extended-spanning-interval)
		     (list s extended-spanning-interval))
		    (second event-occurrence)))))
	 m)
    '()))
  ((contacts?)
   (reduce
    append
    (map (lambda (event-occurrence)
	  (let ((s (term-semiunify (first event-occurrence) e '())))
	   (if (eq? s #f)
	       '()
	       (map (lambda (extended-spanning-interval)
		     (list s extended-spanning-interval))
		    (second event-occurrence)))))
	 m)
    '()))
  ((attached?)
   (reduce
    append
    (map (lambda (event-occurrence)
	  (let ((s (term-semiunify (first event-occurrence) e '())))
	   (if (eq? s #f)
	       '()
	       (map (lambda (extended-spanning-interval)
		     (list s extended-spanning-interval))
		    (second event-occurrence)))))
	 m)
    '()))
  (else
   (panic "Hierarchal event-type references are not (yet) implemented."))))

(define (objects-in-model m)
 (reduce
  unionq
  (map (lambda (event-occurrence)
	(remove-duplicatesq (rest (first event-occurrence))))
       m)
  '()))

(define (occurrences-in-model-some v e m bindings ds)
 (if (null? v)
     (occurrences-in-model-internal e m bindings ds)
     (make-irredundant
      (reduce append
	      (map (lambda (object)
		    (occurrences-in-model-some
		     (rest v) e m (cons (list (first v) object) bindings) ds))
		   (objects-in-model m))
	      '()))))

(define (occurrences-in-model-every-internal v o e m bindings ds)
 (if (null? o)
     (all-intervals)
     (make-irredundant
      (intersection-of
       (occurrences-in-model-internal
	e m (cons (list v (first o)) bindings) ds)
       (occurrences-in-model-every-internal v (rest o) e m bindings ds)))))

(define (occurrences-in-model-every v e m bindings ds)
 (if (null? v)
     (occurrences-in-model-internal e m bindings ds)
     (make-irredundant
      (intersection-of
       (occurrences-in-model-every-internal
	(first v) (objects-in-model m) e m bindings ds)
       (occurrences-in-model-every (rest v) e m bindings ds)))))

(define (occurrences-in-model-and rs es m bindings ds)
 (if (null? (rest es))
     (occurrences-in-model-internal (first es) m bindings ds)
     (make-irredundant
      (reduce
       append
       (map (lambda (r)
	     (span-allen
	      (occurrences-in-model-internal (first es) m bindings ds)
	      r
	      (occurrences-in-model-and rs (rest es) m bindings ds)))
	    rs)
       '()))))

(define (occurrences-in-model-internal e m bindings ds)
 (define (evaluate-arguments e bindings)
  (map (lambda (e) (if (is-variable? e) (second (assq e bindings)) e)) e))
 (define (swap-arguments e) (list (first e) (third e) (second e)))
 (make-irredundant
  (case (first e)
   ((not)
    (complement-of (occurrences-in-model-internal (second e) m bindings ds)))
   ((or)
    (reduce append
	    (map (lambda (e) (occurrences-in-model-internal e m bindings ds))
		 (rest e))
	    '()))
   ((every) (occurrences-in-model-every (second e) (third e) m bindings ds))
   ((some) (occurrences-in-model-some (second e) (third e) m bindings ds))
   ((and)
    (cond
     ((null? (rest e)) (all-intervals))
     ((and (list? (second e))
	   (subsetq? (second e) '(= < > m mi o oi s si f fi d di)))
      (occurrences-in-model-and (second e) (rest (rest e)) m bindings ds))
     (else
      (occurrences-in-model-internal `(and (=) ,@(rest e)) m bindings ds))))
   ((begin)
    (occurrences-in-model-internal `(and (m) ,@(rest e)) m bindings ds))
   ((diamond)
    (if (= (length e) 2)
	(occurrences-in-model-internal
	 `(diamond (= o oi s si f fi d di) ,(second e)) m bindings ds)
	(reduce
	 append
	 (map (lambda (r)
	       (diamond-allen
		r
		(occurrences-in-model-internal (third e) m bindings ds)))
	      (second e))
	 '())))
   ((equal?) (let ((e (evaluate-arguments e bindings)))
	      (if (equal? (second e) (third e)) (all-intervals) '())))
   ((supported?)
    (let ((e (evaluate-arguments e bindings)))
     (map (lambda (event-occurrence) (first (second event-occurrence)))
	  (remove-if-not
	   (lambda (event-occurrence) (equal? (first event-occurrence) e))
	   m))))
   ((supports?)
    (let ((e (evaluate-arguments e bindings)))
     (map (lambda (event-occurrence) (first (second event-occurrence)))
	  (remove-if-not
	   (lambda (event-occurrence) (equal? (first event-occurrence) e))
	   m))))
   ((contacts?)
    (let ((e (evaluate-arguments e bindings)))
     (map (lambda (event-occurrence) (first (second event-occurrence)))
	  (remove-if-not
	   (lambda (event-occurrence)
	    (or (equal? (first event-occurrence) e)
		(equal? (first event-occurrence) (swap-arguments e))))
	   m))))
   ((attached?)
    (let ((e (evaluate-arguments e bindings)))
     (map (lambda (event-occurrence) (first (second event-occurrence)))
	  (remove-if-not
	   (lambda (event-occurrence)
	    (or (equal? (first event-occurrence) e)
		(equal? (first event-occurrence) (swap-arguments e))))
	   m))))
   (else (let* ((d (find-if (lambda (d) (eq? (first e) (first (first d)))) ds))
		(e (evaluate-arguments e bindings)))
	  (occurrences-in-model-internal
	   (second d)
	   m
	   (append (map list (rest (first d)) (rest e)) bindings)
	   ds))))))

(define (occurrences-in-model e m ds)
 (occurrences-in-model-internal e m '() ds))

(define (initial-moment m)
 (let ((start (reduce min
		      (map (lambda (event-occurrence)
			    (extended-spanning-interval-i
			     (first (second event-occurrence))))
			   m)
		      #f)))
  (map first
       (remove-if-not
	(lambda (e)
	 (= start (extended-spanning-interval-i (first (second e)))))
	m))))

(define (final-moment m)
 (let ((finish (reduce max
		       (map (lambda (event-occurrence)
			     (extended-spanning-interval-l
			      (first (second event-occurrence))))
			    m)
		       #f)))
  (map first
       (remove-if-not
        (lambda (e)
         (= finish (extended-spanning-interval-l (first (second e)))))
        m))))

;;; For this comparison to be meaningful we assume that m1 and m2 have
;;; been internalized relative to the target event which is assumed to
;;; be the first event listed in the models.

(define (add-del-list-cmp p? m1 m2 map1 map2)
 (let* ((m11 (model-map-constants m1 map1))
	(m12 (model-map-constants m1 map2))
	(add2 (set-difference (final-moment m2) (initial-moment m2)))
	(del2 (set-difference (initial-moment m2) (final-moment m2)))
	(add11 (set-difference (final-moment m11) (initial-moment m11)))
	(del11 (set-difference (initial-moment m11) (final-moment m11)))
	(add12 (set-difference (final-moment m12) (initial-moment m12)))
	(del12 (set-difference (initial-moment m12) (final-moment m12)))
	(int-num1 (+ (length (intersection add2 add11))
		     (length (intersection del2 del11))))
	(int-num2 (+ (length (intersection add2 add12))
		     (length (intersection del2 del12)))))
  (p? int-num1 int-num2)))

(define (an-object-mapping m1 m2)
 (let ((c1 (constants-in-model m1))
       (c2 (constants-in-model m2)))
  (cond ((< (length c1) (length c2))
	 (map cons (cons 'dummy c1) (a-permutation-of c2)))
	((> (length c1) (length c2))
	 (map cons (a-permutation-of c1) (cons 'dummy c2)))
	(else (map cons (a-permutation-of c1) c2)))))

(define (map-model m1 m2)
 (model-map-constants
  m1
  (first
   (minimal-values (lambda (map1 map2) (add-del-list-cmp > m1 m2 map1 map2))
		   (lambda (map1 map2) (add-del-list-cmp >= m1 m2 map1 map2))
		   (lambda () (an-object-mapping m1 m2))))))

(define (map-models ms)
 (cons (first ms) (map (lambda (m) (map-model m (first ms))) (rest ms))))

(define (ama:classify ds m)
 (reduce
  append
  (map
   (lambda (d)
    (reduce
     append
     (all-values
      (let ((s (map (lambda (argument)
		     (cons argument (a-member-of (objects-in-model m))))
		    (rest (first d)))))
       ;; This forces the substitution to be one-to-one because we don't
       ;; learn (NOT (DIAMOND (EQUAL? X Y))) conjuncts.
       (unless (= (length (remove-duplicatesq (map cdr s))) (length s)) (fail))
       (map (lambda (extended-spaning-interval)
	     (list (ama:apply-substitution s (first d))
		   (list extended-spaning-interval)))
	    (occurrences-in-model (ama:apply-substitution s (first d)) m ds))))
     '()))
   ds)
  '()))

(define (externalize-definition d)
 (let ((variables (unionq (variables-in (first d)) (variables-in (second d)))))
  (list
   (let loop ((t (first d)))
    (cond ((pair? t) (map loop t))
	  ((ama:variable? t)
	   (string->symbol (format "?~s" (positionq t variables))))
	  (else t)))
   (let loop ((t (second d)))
    (cond ((pair? t) (map loop t))
	  ((ama:variable? t)
	   (string->symbol (format "?~s" (positionq t variables))))
	  (else t))))))

(define (lift-definition d)
 (let* ((constants (unionq (constants-in (first d)) (constants-in (second d))))
	(s (map (lambda (constant)
		 (cons constant (create-variable constant constant)))
		constants)))

  (list (ama:apply-substitution s (first d))
	(ama:apply-substitution s (second d)))))

(define (learn k
	       l
	       correspondence?
	       model->ama
	       corpus
	       subject
	       event-type
	       condition
	       training-indices)
 (let ((ms (only-primitive-events
	    (split-occurrences
	     (let ((ms (read-models-of-type-and-condition
			corpus subject event-type condition)))
	      (map (lambda (training-index) (list-ref ms training-index))
		   training-indices))))))
  (lift-definition
   (list
    (second (assoc (list event-type condition) *templates*))
    (k-l-amas-lgg
     k
     l
     (map
      ;; This does not internalize the event because training examples are
      ;; taken as examples only of the primary event type and are assumed to
      ;; already be internal.
      (lambda (m) (k-l-ama-generalization k l (model->ama m)))
      (if correspondence? (map-models ms) ms)))))))

(define (sum-rows m)
  (map-vector (lambda (r) (reduce + (vector->list r) #t)) m))

(define (sum-columns m)
  (sum-rows (transpose m)))

(define (apply-const-subst l subst)
  (cond ((pair? l)
         (cons (apply-const-subst (car l) subst)
               (apply-const-subst (cdr l) subst)))
        (else
         (let ((match (assoc l subst)))
           (if match
               (second match)
               l)))))

(define (new-read-models corpus subject event-type indices)
  (let* ((conditions (map (lambda (i) 
                            (first (num->condition-index i)))
                          indices))
         (offsets (map (lambda (i) 
                         (second (num->condition-index i)))
                       indices))
         (ms (map (lambda (condition index)
                    (if (and (equal? event-type "move")
                             (equal? condition "right"))
                        (apply-const-subst 
                         (read-event-file
                          (event-pathname corpus subject 
                                          event-type condition index))
                         '((green blue) (blue green)))
                        (read-event-file
                         (event-pathname corpus subject 
                                         event-type condition index))))
                  conditions 
                  offsets))
         (constants (reduce append
                            (map constants-in-model ms) #f)))
    (map 
     (lambda (m)
       (map 
        (lambda (e)
          (cons (if (or (equal? 'contacts? (caar e))
                        (equal? 'attached? (caar e)))
                    (let ((a (second (car e)))
                          (b (third (car e))))
                      (if (before-in-list a b constants)
                          (list (caar e) a b)
                          (list (caar e) b a)))
                    (car e))
                (cdr e)))
        m))
     ms)))
       
(define (new-learn k
	           l
	           correspondence?
	           model->ama
                   corpus
	           subject
	           event-type
	           indices)
  (pp indices)
  (newline)
  (let ((ms (only-primitive-events
	    (split-occurrences
	     (new-read-models corpus subject event-type indices)))))
  (lift-definition
   (list
    (second (assoc (list event-type "left") *templates*))
    (k-l-amas-lgg
     k
     l
     (map
      ;; This does not internalize the event because training examples are
      ;; taken as examples only of the primary event type and are assumed to
      ;; already be internal.
      (lambda (m) (k-l-ama-generalization k l (model->ama m)))
      (if correspondence? (map-models ms) ms)))))))

(define (unique-rand-num-list n upper)
  (if (> n upper)
      '()
      (let loop ((l '()))
        (if (= (length l) n)
            l
            (loop (remove-duplicates (cons (random-integer upper) l)))))))

;; trials - number of learning/evaluation trials.
;; sample-size - the size of each training set.
;; event-type - a list giving the strings of the event-types we want to learn.
;; ks - a list giving the k value to use for each event-type.
;; leonard-fp - a file pointer where the Leonard output is written to.
;; def-fp - a file pointer where the learned definitions from each trial are written to.
;; stats-fp - a file pointer where the final results are written to.

(define (random-sample-evaluate trials
                                sample-size
                                event-types
                                ks
	                        l
	                        correspondence?
	                        model->ama
	                        corpus
	                        subject
                                leonard-fp
                                def-fp
                                stats-fp)
  (define (event-type-num e)
    (cond ((equal? e 'pick-up) 0)
          ((equal? e 'put-down) 1)
          ((equal? e 'stack) 2)
          ((equal? e 'unstack) 3)
          ((equal? e 'move) 4)
          ((equal? e 'assemble) 5)
          ((equal? e 'disassemble) 6)
          (else (panic "unknown event-type"))))          
  (define (movie-type-num e)
    (cond ((equal? e "pick-up") 0)
          ((equal? e "put-down") 1)
          ((equal? e "stack") 2)
          ((equal? e "unstack") 3)
          ((equal? e "move") 4)
          ((equal? e "assemble") 5)
          ((equal? e "disassemble") 6)
          (else (panic "unkown movie type"))))
  (set! *lifted?* #f)  
  (let ((trial-results
         (map
          (lambda (T)     
            (let* ((example-num (* (length *conditions*) (length *indices*)))
                   (trn-indices (map (lambda (x) 
                                       (unique-rand-num-list sample-size example-num))
                                     (enumerate (length event-types))))
                   (ds (map (lambda (event-type k indices)                
                              (display (format "Learning ~a~%" event-type))
                              (new-learn k
                                         l
                                         correspondence?
                                         model->ama
                                         corpus
                                         subject
                                         event-type
                                         indices))
                            event-types
                            ks
                            trn-indices))
                   (cm-fp (make-matrix 7 7 0))
                   (cm-fn (make-matrix 7 7 0)))
              (pp (list "Trial" T) def-fp)
	      (newline def-fp)
              (pp (list "Trial" T) leonard-fp)
	      (newline leonard-fp)
              (for-each (lambda (d)
(pp "Printing def . . . ")
                          (pp (externalize-definition d) def-fp)
                          (newline def-fp)
(pp "done.")
(newline))
                        ds)
              (flush-buffer def-fp)
              (for-each
               (lambda (event-type)
                 (for-each
                  (lambda (condition)
                    (for-each
                     (lambda (index)
                       (display (event-pathname corpus subject event-type
                                                condition index) leonard-fp)
                       (newline leonard-fp)
                       (let ((event-occurrences
                              (ama:classify
                               ds
                               (first
                                (only-primitive-events
                                 (split-occurrences
                                  (list
                                   (read-event-file
                                    (event-pathname
                                     corpus subject event-type 
                                     condition index))))))))
                             (correct-results
                              (rest (assoc (list event-type condition)
                                           *correct-results*))))
                         (for-each
                          (lambda (event-occurrence)
                            (write (first event-occurrence) leonard-fp)
                            (display "@" leonard-fp)
                            (display
                             (extended-spanning-interval->string
                              (first (second event-occurrence))) leonard-fp)
                            (newline leonard-fp))
                          event-occurrences)
                         (let* ((results (remove-duplicates 
                                          (map first event-occurrences)))
                                (fp-set (set-difference results correct-results))
                                (fn-set (set-difference correct-results results))
                                (col (movie-type-num event-type)))
                           (for-each
                            (lambda (fp)
			      (let* ((num (event-type-num (first fp)))
				     (val (matrix-ref cm-fp num col)))
				(matrix-set! cm-fp num col
					     (+ 1 val))))
                            fp-set)
                           (for-each 
                            (lambda (fn)
                              (matrix-set! cm-fn (event-type-num (first fn)) col
                                           (+ 1 (matrix-ref 
                                                 cm-fn
                                                 (event-type-num (first fn))
                                                 col))))
                            fn-set)                    
                           (display
                            (format "Correct: ~s~%"
                                    (length (intersection results
                                                          correct-results)))
                            leonard-fp)
                           (display
                            (format "False positives: ~s~%"
                                    (length (set-difference results
                                                            correct-results)))
                            leonard-fp)
                           (display
                            (format "False negatives: ~s~%"
                                    (length (set-difference correct-results
                                                            results)))
                            leonard-fp)
                           (display
                            (format "False positives: ~s, False negatives: ~s~%"
                                    (length (set-difference results
                                                            correct-results))
                                    (length (set-difference correct-results
                                                            results)))
                            leonard-fp)))
                       (newline leonard-fp)
                       (flush-buffer leonard-fp))
                     *indices*))
                  *conditions*))
               *event-types*)
              (pp (list (list "fp-rate" (map 
                                         (lambda (fp fn num-occur)
                                           (/ fp
                                              (- (+ fp num-occur) (+ fn sample-size))))
                                         (vector->list (sum-rows cm-fp))
                                         (vector->list (sum-rows cm-fn))
                                         (map second *num-occurrences*)))
                    (list "fn-rate" (map 
                                         (lambda (fn num-occur)
                                           (/ fn (- num-occur sample-size)))
                                         (vector->list (sum-rows cm-fn))
                                         (map second *num-occurrences*)))
                    (list "errors" (+ (reduce + (vector->list (sum-rows cm-fp)) #t)
                                      (reduce + (vector->list (sum-rows cm-fn)) #t)))
                    (list "fp" (reduce + (vector->list (sum-rows cm-fp)) #t))
                    (list "fn" (reduce + (vector->list (sum-rows cm-fn)) #t))
                    (list "fp-edefs" (sum-rows cm-fp))
                    (list "fn-edefs" (sum-rows cm-fn))-
                    (list "fp-cm" cm-fp) 
                    (list "fn-cm" cm-fn)
                    trn-indices) leonard-fp)
              (newline leonard-fp)
              (flush-buffer leonard-fp)
              (list (list "fp-rate" (map 
                                         (lambda (fp fn num-occur)
                                           (/ fp
                                              (- (+ fp num-occur) (+ fn sample-size))))
                                         (vector->list (sum-rows cm-fp))
                                         (vector->list (sum-rows cm-fn))
                                         (map second *num-occurrences*)))
                    (list "fn-rate" (map 
                                         (lambda (fn num-occur)
                                           (/ fn (- num-occur sample-size)))
                                         (vector->list (sum-rows cm-fn))
                                         (map second *num-occurrences*)))
                    (list "errors" (+ (reduce + (vector->list (sum-rows cm-fp)) #t)
                                      (reduce + (vector->list (sum-rows cm-fn)) #t)))
                    (list "fp" (reduce + (vector->list (sum-rows cm-fp)) #t))
                    (list "fn" (reduce + (vector->list (sum-rows cm-fn)) #t))
                    (list "fp-edefs" (sum-rows cm-fp))
                    (list "fn-edefs" (sum-rows cm-fn))
                    (list "fp-cm" cm-fp) 
                    (list "fn-cm" cm-fn)
                    trn-indices)))     
          (enumerate trials))))
    (pp (list 
     (list "avg-fp-rates"
           (map (lambda (fpr) (/ fpr trials))
                (reduce (lambda (fpr1 fpr2) (map + fpr1 fpr2)) 
                        (map (lambda (t) (second (first t))) trial-results) #f)))
     (list "avg-fn-rates"
           (map (lambda (fnr) (/ fnr trials))
                (reduce (lambda (fnr1 fnr2) (map + fnr1 fnr2)) 
                        (map (lambda (t) (second (second t))) trial-results) #f)))
     trial-results)
     stats-fp)))

(define (cross-validate-event-type folds
                                   k
                                   l
                                   correspondence?
                                   model->ama
                                   corpus
                                   subject
                                   event-type
                                   leonard-fp)
  (let* ((example-num (* (length *conditions*) (length *indices*)))
         (test-num (round (/ example-num folds)))
         (fp 0)
         (fn 0))
    (let loop ((untested (enumerate example-num)))
      (unless (< (length untested) test-num)
        (let* ((test-set (map (lambda (i) (list-ref untested i))
                              (unique-rand-num-list test-num (length untested))))
               (def (list (new-learn k
                                     l
                                     correspondence?
                                     model->ama
                                     corpus
                                     subject
                                     event-type
                                     (set-difference (enumerate example-num)
                                                     test-set)))))
          (newline leonard-fp)
          (pp (list "Test Set" test-set) leonard-fp)
          (pp (list "# of Time Lines" (- (length (second (car def))) 1)) leonard-fp)
          (newline leonard-fp)
          (for-each
           (lambda (condition)
             (for-each
              (lambda (index)
                (display (event-pathname corpus subject event-type
                                         condition index) leonard-fp)
                (newline leonard-fp)
                (let ((event-occurrences
                       (ama:classify
                        def
                        (first
                         (only-primitive-events
                          (split-occurrences
                           (list
                            (read-event-file
                             (event-pathname
                              corpus subject event-type 
                              condition index))))))))
                      (correct-results
                       (rest (assoc (list event-type condition)
                                    *correct-diagonal-results*))))
                  (for-each
                   (lambda (event-occurrence)
                     (write (first event-occurrence) leonard-fp)
                     (display "@" leonard-fp)
                     (display
                      (extended-spanning-interval->string
                       (first (second event-occurrence))) leonard-fp)
                     (newline leonard-fp))
                   event-occurrences)
                  (let* ((results (remove-duplicates 
                                   (map first event-occurrences)))
                         (fp-set (set-difference results 
                                                 correct-results))
                         (fn-set (set-difference correct-results 
                                                 results)))
                    (set! fp (+ fp (length fp-set)))
                    (set! fn (+ fn (length fn-set)))                    
                    (display
                     (format "Correct: ~s~%"
                             (length (intersection results
                                                   correct-results)))
                     leonard-fp)
                    (display
                     (format "False positives: ~s~%"
                             (length (set-difference 
                                      results
                                      correct-results))) leonard-fp)
                    (display
                     (format "False negatives: ~s~%"
                             (length (set-difference correct-results
                                                     results)))
                     leonard-fp)
                    (display
                     (format "False positives: ~s, False negatives: ~s~%"
                             (length (set-difference 
                                      results
                                      correct-results))
                             (length (set-difference correct-results
                                                     results)))
                     leonard-fp))))
              (map (lambda (i) (second (num->condition-index i)))
                   test-set)))
           (map (lambda (i) (first (num->condition-index i)))
                test-set))
      (flush-buffer leonard-fp)
      (loop (set-difference untested test-set)))))
    (list fp fn)))
       
;; There is some junk below, this was borrowed from the CV code.
(define (learn-definitions folds
                    event-types
                    ks
		    l
		    correspondence?
		    model->ama
		    corpus
		    subject
		    leonard-fp
		    def-fp
		    stats-fp)
  (define (event-type-num e)
    (cond ((equal? e 'pick-up) 0)
          ((equal? e 'put-down) 1)
          ((equal? e 'stack) 2)
          ((equal? e 'unstack) 3)
          ((equal? e 'move) 4)
          ((equal? e 'assemble) 5)
          ((equal? e 'disassemble) 6)
          (else (panic "unknown event-type"))))          
  (define (movie-type-num e)
    (cond ((equal? e "pick-up") 0)
          ((equal? e "put-down") 1)
          ((equal? e "stack") 2)
          ((equal? e "unstack") 3)
          ((equal? e "move") 4)
          ((equal? e "assemble") 5)
          ((equal? e "disassemble") 6)
          (else (panic "unkown movie type"))))
  (set! *lifted?* #f)
  (let ((ds (map (lambda (event-type k)                
                   (display (format "Learning ~a~%" event-type))
                   (new-learn k
                              l
                              correspondence?
                              model->ama
                              corpus
                              subject
                              event-type
                              (enumerate (* (length *indices*) (length *conditions*)))))
                 event-types
                 ks)))
    (pp "All Data:" def-fp)
    (newline def-fp)
    (for-each (lambda (d)
                (pp (externalize-definition d) def-fp)
                (newline def-fp))
              ds))
  (flush-buffer def-fp))
    

(define (cross-validation folds
                          event-types
                          ks
                          l
                          correspondence?
                          model->ama
                          corpus
                          subject
                          leonard-fp
                          def-fp
                          stats-fp)
  (define (event-type-num e)
    (cond ((equal? e 'pick-up) 0)
          ((equal? e 'put-down) 1)
          ((equal? e 'stack) 2)
          ((equal? e 'unstack) 3)
          ((equal? e 'move) 4)
          ((equal? e 'assemble) 5)
          ((equal? e 'disassemble) 6)
          (else (panic "unknown event-type"))))          
  (define (movie-type-num e)
    (cond ((equal? e "pick-up") 0)
          ((equal? e "put-down") 1)
          ((equal? e "stack") 2)
          ((equal? e "unstack") 3)
          ((equal? e "move") 4)
          ((equal? e "assemble") 5)
          ((equal? e "disassemble") 6)
          (else (panic "unkown movie type"))))
  (set! *lifted?* #f)
  (let ((cm-fp (make-matrix 7 7 0))
        (cm-fn (make-matrix 7 7 0))
        (ds (map (lambda (event-type k)                
                   (display (format "Learning ~a~%" event-type))
                   (new-learn k
                              l
                              correspondence?
                              model->ama
                              corpus
                              subject
                              event-type
                              (enumerate (* (length *indices*) (length *conditions*)))))
                 event-types
                 ks)))
    (pp "All Data:" def-fp)
    (newline def-fp)
    (pp "All Data:" leonard-fp)
    (newline leonard-fp)
    (for-each (lambda (d)
                (pp (externalize-definition d) def-fp)
                (newline def-fp))
              ds)
    
    ;; Calculate off-diagnal statistics.
    (for-each
     (lambda (event-type)
       (for-each
        (lambda (condition)
          (for-each
           (lambda (index)
             (display (event-pathname corpus subject event-type
                                      condition index) leonard-fp)
             (newline leonard-fp)
             (let ((event-occurrences
                    (ama:classify
                     ds
                     (first
                      (only-primitive-events
                       (split-occurrences
                        (list
                         (read-event-file
                          (event-pathname
                           corpus subject event-type 
                           condition index))))))))
                   (correct-results
                    (rest (assoc (list event-type condition)
                                 *correct-results*))))
               (for-each
                (lambda (event-occurrence)
                  (write (first event-occurrence) leonard-fp)
                  (display "@" leonard-fp)
                  (display
                   (extended-spanning-interval->string
                    (first (second event-occurrence))) leonard-fp)
                  (newline leonard-fp))
                event-occurrences)
               (let* ((results (remove-duplicates 
                                (map first event-occurrences)))
                      (fp-set (set-difference results 
                                              correct-results))
                      (fn-set (set-difference correct-results 
                                              results))
                      (col (movie-type-num event-type)))
                 (for-each
                  (lambda (fp)
                    (matrix-set! cm-fp (event-type-num (first 
                                                        fp)) col
                                                        (+ 1 (matrix-ref 
                                                              cm-fp 
                                                              (event-type-num (first fp))
                                                              col))))
                  fp-set)
                 (for-each 
                  (lambda (fn)
                    (matrix-set! cm-fn (event-type-num (first 
                                                        fn)) col
                                                        (+ 1 (matrix-ref 
                                                              cm-fn
                                                              (event-type-num (first fn))
                                                              col))))
                  fn-set)                    
                 (display
                  (format "Correct: ~s~%"
                          (length (intersection results
                                                correct-results)))
                  leonard-fp)
                 (display
                  (format "False positives: ~s~%"
                          (length (set-difference 
                                   results
                                   correct-results))) leonard-fp)
                 (display
                  (format "False negatives: ~s~%"
                          (length (set-difference correct-results
                                                  results)))
                  leonard-fp)
                 (display
                  (format "False positives: ~s, False negatives: ~s~%"
                          (length (set-difference 
                                   results
                                   correct-results))
                          (length (set-difference correct-results
                                                  results)))
                  leonard-fp)))
             (newline leonard-fp)
             (flush-buffer leonard-fp))
           *indices*))
        *conditions*))
     *event-types*)
    (pp (list "All Data" 
              (list "fp-rate" (map 
                               (lambda (fp fn num-occur)
                                 (/ fp
                                    (- (+ fp num-occur) fn)))
                               (vector->list (sum-rows cm-fp))
                               (vector->list (sum-rows cm-fn))
                               (map second *num-occurrences*)))
              (list "fn-rate" (map 
                               (lambda (fn num-occur)
                                 (/ fn num-occur))
                               (vector->list (sum-rows cm-fn))
                               (map second *num-occurrences*)))
              (list "errors" (+
                              (reduce 
                               + 
                               (vector->list (sum-rows cm-fp)) #t)
                              (reduce + (vector->list 
                                         (sum-rows cm-fn)) #t)))
              (list "fp" (reduce + (vector->list 
                                    (sum-rows cm-fp)) #t))
              (list "fn" (reduce + (vector->list 
                                    (sum-rows cm-fn)) #t))
              (list "fp-edefs" (sum-rows cm-fp))
              (list "fn-edefs" (sum-rows cm-fn))
              (list "fp-cm" cm-fp) 
              (list "fn-cm" cm-fn))
        stats-fp)     
    (flush-buffer stats-fp)
    ;; Calculate diagonal entries (via cross-validation).
    (for-each
     (lambda (event-type k)
       (let ((fp-fn (cross-validate-event-type
                     folds
                     k
                     l
                     correspondence?
                     model->ama
                     corpus
                     subject
                     event-type
                     leonard-fp)))
         (matrix-set! cm-fp (movie-type-num event-type) (movie-type-num event-type) 
                      (first fp-fn)) 
         (matrix-set! cm-fn (movie-type-num event-type) (movie-type-num event-type) 
                      (second fp-fn))))
     event-types
     ks)
    (newline stats-fp)
    (pp (list (list "Hold-Out" folds)
              (list "fp-rate" (map 
                               (lambda (fp fn num-occur)
                                 (/ fp
                                    (- (+ fp num-occur) fn)))
                               (vector->list (sum-rows cm-fp))
                               (vector->list (sum-rows cm-fn))
                               (map second *num-occurrences*)))
              (list "fn-rate" (map 
                               (lambda (fn num-occur)
                                 (/ fn num-occur))
                               (vector->list (sum-rows cm-fn))
                               (map second *num-occurrences*)))
              (list "errors" (+
                              (reduce 
                               + 
                               (vector->list (sum-rows cm-fp)) #t)
                              (reduce + (vector->list 
                                         (sum-rows cm-fn)) #t)))
              (list "fp" (reduce + (vector->list 
                                    (sum-rows cm-fp)) #t))
              (list "fn" (reduce + (vector->list 
                                    (sum-rows cm-fn)) #t))
              (list "fp-edefs" (sum-rows cm-fp))
              (list "fn-edefs" (sum-rows cm-fn))
              (list "fp-cm" cm-fp) 
              (list "fn-cm" cm-fn))
        stats-fp)))

(define (leave-one-out-sample-event-type 
	                           sample-size
                                   k
                                   l
                                   correspondence?
                                   model->ama
                                   corpus
                                   subject
                                   event-type
                                   leonard-fp)
  (let* ((example-num (* (length *conditions*) (length *indices*)))
	 (test-num 1)
         (fp 0)
         (fn 0))
    (let loop ((untested (enumerate example-num)))
      (unless (< (length untested) test-num)
        (let* ((test-set (map (lambda (i) (list-ref untested i))
                              (unique-rand-num-list test-num (length untested))))
	       (available-train (set-difference (enumerate example-num) test-set))
	       (trn-set (map (lambda (i) (list-ref available-train i))
			      (unique-rand-num-list sample-size (length available-train)))) 
               (def (list (new-learn k
                                     l
                                     correspondence?
                                     model->ama
                                     corpus
                                     subject
                                     event-type
                                     trn-set))))
          (newline leonard-fp)
          (pp (list "Train Set" trn-set) leonard-fp)
          (pp (list "# of Time Lines" (- (length (second (car def))) 1)) leonard-fp)
          (newline leonard-fp)
          (for-each
           (lambda (condition)
             (for-each
              (lambda (index)
                (display (event-pathname corpus subject event-type
                                         condition index) leonard-fp)
                (newline leonard-fp)
                (let ((event-occurrences
                       (ama:classify
                        def
                        (first
                         (only-primitive-events
                          (split-occurrences
                           (list
                            (read-event-file
                             (event-pathname
                              corpus subject event-type 
                              condition index))))))))
                      (correct-results
                       (rest (assoc (list event-type condition)
                                    *correct-diagonal-results*))))
                  (for-each
                   (lambda (event-occurrence)
                     (write (first event-occurrence) leonard-fp)
                     (display "@" leonard-fp)
                     (display
                      (extended-spanning-interval->string
                       (first (second event-occurrence))) leonard-fp)
                     (newline leonard-fp))
                   event-occurrences)
                  (let* ((results (remove-duplicates 
                                   (map first event-occurrences)))
                         (fp-set (set-difference results 
                                                 correct-results))
                         (fn-set (set-difference correct-results 
                                                 results)))
                    (set! fp (+ fp (length fp-set)))
                    (set! fn (+ fn (length fn-set)))                    
                    (display
                     (format "Correct: ~s~%"
                             (length (intersection results
                                                   correct-results)))
                     leonard-fp)
                    (display
                     (format "False positives: ~s~%"
                             (length (set-difference 
                                      results
                                      correct-results))) leonard-fp)
                    (display
                     (format "False negatives: ~s~%"
                             (length (set-difference correct-results
                                                     results)))
                     leonard-fp)
                    (display
                     (format "False positives: ~s, False negatives: ~s~%"
                             (length (set-difference 
                                      results
                                      correct-results))
                             (length (set-difference correct-results
                                                     results)))
                     leonard-fp))))
              (map (lambda (i) (second (num->condition-index i)))
                   test-set)))
           (map (lambda (i) (first (num->condition-index i)))
                test-set))
      (flush-buffer leonard-fp)
      (loop (set-difference untested test-set)))))
    (list fp fn)))
       
(define (leave-one-out-sample sample-size
			      event-types
			      ks
			      l
			      correspondence?
			      model->ama
			      corpus
			      subject
			      leonard-fp)
  (define (event-type-num e)
    (cond ((equal? e 'pick-up) 0)
          ((equal? e 'put-down) 1)
          ((equal? e 'stack) 2)
          ((equal? e 'unstack) 3)
          ((equal? e 'move) 4)
          ((equal? e 'assemble) 5)
          ((equal? e 'disassemble) 6)
          (else (panic "unknown event-type"))))          
  (define (movie-type-num e)
    (cond ((equal? e "pick-up") 0)
          ((equal? e "put-down") 1)
          ((equal? e "stack") 2)
          ((equal? e "unstack") 3)
          ((equal? e "move") 4)
          ((equal? e "assemble") 5)
          ((equal? e "disassemble") 6)
          (else (panic "unkown movie type"))))
  (set! *lifted?* #f)
  (let* ((cm-fp (make-matrix 7 7 0))
	 (cm-fn (make-matrix 7 7 0))
	 (trn-indices (map (lambda (x) 
			     (unique-rand-num-list sample-size 
						   (* (length *indices*) (length *conditions*))))
			   (enumerate (length event-types))))
	 (ds (map (lambda (event-type k indices)                
		    (display (format "Learning ~a~%" event-type))
		    (new-learn k
			       l
			       correspondence?
			       model->ama
			       corpus
			       subject
			       event-type
			       indices))
		  event-types
		  ks
		  trn-indices)))
    (pp "Off Diagonal:" leonard-fp)
    (newline leonard-fp)
    (pp trn-indices leonard-fp)
    (newline leonard-fp)
    
    ;; Calculate off-diagnal statistics.
    (for-each
     (lambda (event-type)
       (for-each
        (lambda (condition)
          (for-each
           (lambda (index)
             (display (event-pathname corpus subject event-type
                                      condition index) leonard-fp)
             (newline leonard-fp)
             (let ((event-occurrences
                    (ama:classify
                     ds
                     (first
                      (only-primitive-events
                       (split-occurrences
                        (list
                         (read-event-file
                          (event-pathname
                           corpus subject event-type 
                           condition index))))))))
                   (correct-results
                    (rest (assoc (list event-type condition)
                                 *correct-results*))))
               (for-each
                (lambda (event-occurrence)
                  (write (first event-occurrence) leonard-fp)
                  (display "@" leonard-fp)
                  (display
                   (extended-spanning-interval->string
                    (first (second event-occurrence))) leonard-fp)
                  (newline leonard-fp))
                event-occurrences)
               (let* ((results (remove-duplicates 
                                (map first event-occurrences)))
                      (fp-set (set-difference results 
                                              correct-results))
                      (fn-set (set-difference correct-results 
                                              results))
                      (col (movie-type-num event-type)))
                 (for-each
                  (lambda (fp)
                    (matrix-set! cm-fp (event-type-num (first 
                                                        fp)) col
                                                        (+ 1 (matrix-ref 
                                                              cm-fp 
                                                              (event-type-num (first fp))
                                                              col))))
                  fp-set)
                 (for-each 
                  (lambda (fn)
                    (matrix-set! cm-fn (event-type-num (first 
                                                        fn)) col
                                                        (+ 1 (matrix-ref 
                                                              cm-fn
                                                              (event-type-num (first fn))
                                                              col))))
                  fn-set)                    
                 (display
                  (format "Correct: ~s~%"
                          (length (intersection results
                                                correct-results)))
                  leonard-fp)
                 (display
                  (format "False positives: ~s~%"
                          (length (set-difference 
                                   results
                                   correct-results))) leonard-fp)
                 (display
                  (format "False negatives: ~s~%"
                          (length (set-difference correct-results
                                                  results)))
                  leonard-fp)
                 (display
                  (format "False positives: ~s, False negatives: ~s~%"
                          (length (set-difference 
                                   results
                                   correct-results))
                          (length (set-difference correct-results
                                                  results)))
                  leonard-fp)))
             (newline leonard-fp)
             (flush-buffer leonard-fp))
           *indices*))
        *conditions*))
     *event-types*)
    ;; Calculate diagonal entries (via cross-validation).
    (for-each
     (lambda (event-type k)
       (let ((fp-fn (leave-one-out-sample-event-type
                     sample-size
                     k
                     l
                     correspondence?
                     model->ama
                     corpus
                     subject
                     event-type
                     leonard-fp)))
         (matrix-set! cm-fp (movie-type-num event-type) (movie-type-num event-type) 
                      (first fp-fn)) 
         (matrix-set! cm-fn (movie-type-num event-type) (movie-type-num event-type) 
                      (second fp-fn))))
     event-types
     ks)
    (list (list "fp-rate" (map 
                               (lambda (fp fn num-occur)
                                 (/ fp
                                    (- (+ fp num-occur) fn)))
                               (vector->list (sum-rows cm-fp))
                               (vector->list (sum-rows cm-fn))
                               (map second *num-occurrences*)))
              (list "fn-rate" (map 
                               (lambda (fn num-occur)
                                 (/ fn num-occur))
                               (vector->list (sum-rows cm-fn))
                               (map second *num-occurrences*)))
              (list "errors" (+
                              (reduce 
                               + 
                               (vector->list (sum-rows cm-fp)) #t)
                              (reduce + (vector->list 
                                         (sum-rows cm-fn)) #t)))
              (list "fp" (reduce + (vector->list 
                                    (sum-rows cm-fp)) #t))
              (list "fn" (reduce + (vector->list 
                                    (sum-rows cm-fn)) #t))
              (list "fp-edefs" (sum-rows cm-fp))
              (list "fn-edefs" (sum-rows cm-fn))
              (list "fp-cm" cm-fp) 
              (list "fn-cm" cm-fn))))

(define (test1 k l event-type)
 (set! *lifted?* #t)
 (pp (externalize-term
      (k-l-amas-lgg
       k
       l
       (map (lambda (m)
	     (k-l-ama-generalization k l (model->ama (internalize-model m))))
	    (models-for event-type)))))
 (newline))

(define (test2 k l event-type)
 (set! *lifted?* #t)
 (let* ((e (k-l-amas-lgg
	    k
	    l
	    (map
	     (lambda (m)
	      (k-l-ama-generalization k l (model->ama (internalize-model m))))
	     (models-for event-type))))
	(ms (read-models))
	(positive-models
	 (remove-if-not
	  (lambda (m)
	   (some (lambda (event-occurrence)
		  (eq? (first (first event-occurrence)) event-type))
		 m))
	  ms))
	(negative-models
	 (remove-if
	  (lambda (m)
	   (some (lambda (event-occurrence)
		  (eq? (first (first event-occurrence)) event-type))
		 m))
	  ms)))
  (display "False negatives: ")
  (write (- (length positive-models)
	    (count-if (lambda (m)
		       (not
			(null?
			 (fast-occurrences-in-model
			  e (remove-if-not primitive-event-occurrence? m)))))
		      positive-models)))
  (display "/")
  (write (length positive-models))
  (newline)
  (display "False positives: ")
  (write (count-if (lambda (m)
		    (not
		     (null?
		      (fast-occurrences-in-model
		       e (remove-if-not primitive-event-occurrence? m)))))
		   negative-models))
  (display "/")
  (write (length negative-models))
  (newline)))

(define (test3 e)
 (define (print-occurrence occurrence)
  (when (= (length (map cdr (first occurrence)))
	   (length (remove-duplicates (map cdr (first occurrence)))))
   (write (first occurrence))
   (display " ")
   (display (extended-spanning-interval->string (second occurrence)))
   (newline)))
 (set! *lifted?* #t)
 (for-each
  (lambda (m)
   (for-each
    (lambda (event-occurrence)
     (unless (primitive-event-occurrence? event-occurrence)
      (write (first event-occurrence))
      (for-each
       (lambda (extended-spanning-interval)
	(display " ")
	(display
	 (extended-spanning-interval->string extended-spanning-interval)))
       (second event-occurrence))
      (newline)))
    m)
   (for-each print-occurrence (fast-occurrences-in-model e m))
   (newline))
  (read-models)))

(define (test4 i e)
 (define (print-occurrence occurrence)
  (when (= (length (map cdr (first occurrence)))
	   (length (remove-duplicates (map cdr (first occurrence)))))
   (write (first occurrence))
   (display " ")
   (display (extended-spanning-interval->string (second occurrence)))
   (newline)))
 (set! *lifted?* #t)
 (let ((m (list-ref (read-models) i)))
  (for-each
   (lambda (event-occurrence)
    (write (first event-occurrence))
    (for-each
     (lambda (extended-spanning-interval)
      (display " ")
      (display
       (extended-spanning-interval->string extended-spanning-interval)))
     (second event-occurrence))
    (newline))
   m)
  (for-each print-occurrence (fast-occurrences-in-model e m))))

(define (test5 ks
	       l
	       correspondence?
	       model->ama
	       corpus
	       subject
	       condition
	       training-indices)
 (set! *lifted?* #f)
 (let ((ds (map (lambda (event-type k)
		 (display (format "Learning ~a~%" event-type)) ;debugging
		 (learn k
			l
			correspondence?
			model->ama
			corpus
			subject
			event-type
			condition
			training-indices))
		*event-types*
		ks)))
  (for-each (lambda (d)
	     (pp (externalize-definition d))
	     (newline))
	    ds)
  (for-each
   (lambda (event-type)
    (for-each
     (lambda (condition)
      (for-each
       (lambda (index)
	(display (event-pathname corpus subject event-type condition index))
	(newline)
	(let ((event-occurrences
	       (ama:classify
		ds
		(first
		 (only-primitive-events
		  (split-occurrences
		   (list
		    (read-event-file
		     (event-pathname
		      corpus subject event-type condition index))))))))
	      (correct-results
	       (rest (assoc (list event-type condition) *correct-results*))))
	 (for-each
	  (lambda (event-occurrence)
	   (write (first event-occurrence))
	   (display "@")
	   (display
	    (extended-spanning-interval->string
	     (first (second event-occurrence))))
	   (newline))
	  event-occurrences)
	 (let ((results (remove-duplicates (map first event-occurrences))))
	  (display
	   (format "Correct: ~s~%"
		   (length (intersection results correct-results))))
	  (display
	   (format "False positives: ~s~%"
		   (length (set-difference results correct-results))))
	  (display
	   (format "False negatives: ~s~%"
		   (length (set-difference correct-results results))))
	  (display
	   (format "False positives: ~s, False negatives: ~s~%"
		   (length (set-difference results correct-results))
		   (length (set-difference correct-results results))))))
	(newline))
       *indices*))
     *conditions*))
   *event-types*)))

(define (test6 corpus subject ds)
 (for-each
  (lambda (event-type)
   (for-each
    (lambda (condition)
     (for-each
      (lambda (index)
       (display (event-pathname corpus subject event-type condition index))
       (newline)
       (let ((event-occurrences
	      (ama:classify
	       ds
	       (first
		(only-primitive-events
		 (split-occurrences
		  (list
		   (read-event-file
		    (event-pathname
		     corpus subject event-type condition index))))))))
	     (correct-results
	      (rest (assoc (list event-type condition) *correct-results*))))
	(for-each
	 (lambda (event-occurrence)
	  (write (first event-occurrence))
	  (display "@")
	  (display
	   (extended-spanning-interval->string
	    (first (second event-occurrence))))
	  (newline))
	 event-occurrences)
	(let ((results (remove-duplicates (map first event-occurrences))))
	 (display
	  (format "Correct: ~s~%"
		  (length (intersection results correct-results))))
	 (display
	  (format "False positives: ~s~%"
		  (length (set-difference results correct-results))))
	 (display
	  (format "False negatives: ~s~%"
		  (length (set-difference correct-results results))))
	 (display
	  (format "False positives: ~s, False negatives: ~s~%"
		  (length (set-difference results correct-results))
		  (length (set-difference correct-results results))))))
       (newline))
      *indices*))
    *conditions*))
  *event-types*))

(define (test7 trials
               sample-size
	       ks
	       l
	       correspondence?
	       model->ama
	       corpus
	       subject
	       basename)
  (let ((leonard-fp (open-output-file 
		     (string-append basename ".leonard")))
	(def-fp (open-output-file
		 (string-append basename ".definitions")))
	(stats-fp (open-output-file
		   (string-append basename ".stats")))) 
    (random-sample-evaluate trials
			    sample-size
			    *event-types*
			    ks
			    l
			    correspondence?
			    model->ama
			    corpus
			    subject
			    leonard-fp
			    def-fp
			    stats-fp)))

(define (test8 folds
               ks
	       l
	       correspondence?
	       model->ama
	       corpus
	       subject
	       basename)
  (let ((leonard-fp (open-output-file 
		     (string-append basename ".leonard")))
	(def-fp (open-output-file
		 (string-append basename ".definitions")))
	(stats-fp (open-output-file
		   (string-append basename ".stats")))) 
    (cross-validation folds
		      *event-types*
		      ks
		      l
		      correspondence?
		      model->ama
		      corpus
		      subject
		      leonard-fp
		      def-fp
		      stats-fp)))

(define (test10 folds
		ks
		l
		correspondence?
		model->ama
		corpus
		subject
		basename)
  (let ((def-fp (open-output-file
	 (string-append basename ".definitions"))))
    (learn-definitions folds
		*event-types*
		ks
		l
		correspondence?
		model->ama
		corpus
		subject
		def-fp
		def-fp
		def-fp)))

(define (test9 trials
	       sample-size
               ks
	       l
	       correspondence?
	       model->ama
	       corpus
	       subject
	       basename)
  (let* ((leonard-fp (open-output-file 
		     (string-append basename ".leonard")))
	 (stats-fp (open-output-file
		    (string-append basename ".stats")))
	 (trial-results 
	  (map (lambda (T)
		 (pp (list "Trial" T) leonard-fp)
		 (newline leonard-fp)
		 (pp (list "Trial" T))
		 (newline)
		 (let ((result (leave-one-out-sample sample-size
						     *event-types*
						     ks
						     l
						     correspondence?
						     model->ama
						     corpus
						     subject
						     leonard-fp)))
		   (pp result stats-fp)
		   (newline stats-fp)
                   (flush-buffer stats-fp)
		   result))
	       (enumerate trials))))
        (pp 
	 (list 
	  (list "Sample-Size" sample-size)
	  (list "Trials" trials)
	  (list "avg-fp-rates"
		(map (lambda (fpr) (/ fpr trials))
		     (reduce (lambda (fpr1 fpr2) (map + fpr1 fpr2)) 
			     (map (lambda (t) (second (first t))) trial-results) #f)))
	  (list "avg-fn-rates"
		(map (lambda (fnr) (/ fnr trials))
		     (reduce (lambda (fnr1 fnr2) (map + fnr1 fnr2)) 
			     (map (lambda (t) (second (second t))) trial-results) #f)))
	  (list "avg-fp-num"
		(map (lambda (fp) (/ fp trials))
		     (reduce (lambda (fp1 fp2) (map + fp1 fp2)) 
			     (map (lambda (t) (vector->list (second (list-ref t 5)))) trial-results) #f)))
	  (list "avg-fn-num"
		(map (lambda (fn) (/ fn trials))
		     (reduce (lambda (fn1 fn2) (map + fn1 fn2)) 
			     (map (lambda (t) (vector->list (second (list-ref t 6)))) trial-results) #f))))
	 stats-fp)
	(newline stats-fp)))
 

(define (test-add-del-list event-type)
 (let* ((models (map (lambda (m) (internalize-model m))
		     (models-for event-type))))
  (for-each
   (lambda (m1)
    (for-each
     (lambda (m2)
      (unless (eq? m2 m1)
       (pp (first (first m1)))
       (newline)
       (pp (first (first m2)))
       (newline)
       (pp (minimal-values (lambda (map1 map2)
			    (add-del-list-cmp > m1 m2 map1 map2))
			   (lambda (map1 map2)
			    (add-del-list-cmp >= m1 m2 map1 map2))
			   (lambda () (an-object-mapping m1 m2))))
       (newline)
       (newline)))
     models))
   models)))

;;; Top Level

(define-command (main
		 (at-most-one ("corpus"
			       corpus?
			       (corpus "string" string-argument "w28nov2001")))
		 (at-most-one ("data"
			       data?
			       (data "string" string-argument "./data-m19nov2001")))
		 (at-most-one ("subject"
			       subject?
			       (subject "string" string-argument "qobi")))
		 (at-most-one
		  ("definitions"
		   definitions?
		   (definitions "pathname" string-argument "default")))
                 (at-most-one 
                  ("rnd-sample"
                   rnd-sample?
                   (rnd-sample "pathname" string-argument "random-sample")))
		 (at-most-one
		  ("sample-l1o"
		   sample-l1o?
		   (sample-l1o "pathname" string-argument "sample-leave-one-out")))
		 (at-most-one
		  ("cv"
		   cv?
		   (cv "pathname" string-argument "cross-validation")))
		 (at-most-one
		  ("learn-defs"
		   learn-defs?
		   (learn-defs "pathname" string-argument
"learn-definitions")))
                 (at-most-one
                  ("sample-size"
                   sample-size?
                   (sample-size "#" integer-argument 10)))                   
                 (at-most-one
                  ("trials"
                   trials?
                   (trials "#" integer-argument 5)))
		 (at-most-one
		  ("folds"
		   folds?
		   (folds "#" integer-argument 10)))
		 (any-number ("k" k? (ks "#" integer-argument)))
		 (at-most-one ("l" l? (l "#" integer-argument #f)))
		 (at-most-one ("correspondence" correspondence?))
		 (at-most-one ("p" p?)
			      ("n" n?)
			      ("bn" bn?))
		 (at-most-one ("left" left?)
			      ("right" right?))
		 (any-number ("i" i? (training-indices "#" integer-argument))))
 (cond
  (definitions?
   (when (or k? l? correspondence? p? n? left? right? i?)
	 (panic "Invalid option with -definitions"))
   (set! *howard-directory* data)
   (test6 corpus 
	  subject
	  (read-object-from-file definitions)))
  (rnd-sample?
;;   (randomize)
   (test7 trials
          sample-size
          (reverse ks)
          l
          correspondence?
	  (cond (p? model->ama)
		(n? partial-model->ama-n)
		(bn? partial-model->ama-pre-post-n)
		(else (fuck-up)))
	  corpus
	  subject
          rnd-sample))
  (cv?
;;   (randomize)
   (set! *howard-directory* data)
   (test8 folds
          (reverse ks)
          l
          correspondence?
	  (cond (p? model->ama)
		(n? partial-model->ama-n)
		(bn? partial-model->ama-pre-post-n)
		(else (fuck-up)))
	  corpus
	  subject
          cv))
  (learn-defs?
;;   (randomize)
   (set! *howard-directory* data)
   (test10 30
          (reverse ks)
          l
          correspondence?
	  (cond (p? model->ama)
		(n? partial-model->ama-n)
		(bn? partial-model->ama-pre-post-n)
		(else (fuck-up)))
	  corpus
	  subject
          learn-defs))
  (sample-l1o?
;;   (randomize)
   (set! *howard-directory* data)
   (test9 trials
          sample-size
          (reverse ks)
          l
          correspondence?
	  (cond (p? model->ama)
		(n? partial-model->ama-n)
		(bn? partial-model->ama-pre-post-n)
		(else (fuck-up)))
	  corpus
	  subject
          sample-l1o))
  (else
   (unless (or p? n? bn?)
    (panic "You must specify a negation type either -p, -bn, or -n"))
   (unless (or left? right?)
    (panic "You must specify either -left or -right"))
   (test5 (reverse ks)
	  l
	  correspondence?
	  (cond (p? model->ama)
		(n? partial-model->ama-n)
		(bn? partial-model->ama-pre-post-n)
		(else (fuck-up)))
	  corpus
	  subject
	  (cond (left? "left")
		(right? "right")
		(else (fuck-up)))
	  (reverse training-indices)))))

;;; Tam V'Nishlam Shevah L'El Borei Olam













