;;; Procedures from HLIB-SPAN

(define *fuzz* 0)

(define-structure extended-spanning-interval
 i j k l alpha? beta? gamma? delta? epsilon? zeta?)

(define (implies p q) (or (not p) q))

(define (all-intervals)
 (append (spanning-intervals minus-infinity infinity #f #f #t #t)
	 (spanning-intervals minus-infinity infinity #f #t #t #t)
	 (spanning-intervals minus-infinity infinity #t #f #t #t)
	 (spanning-intervals minus-infinity infinity #t #t #t #t)))

(define (extended-spanning-intervals i
				     j
				     k
				     l
				     alpha?
				     beta?
				     gamma?
				     delta?
				     epsilon?
				     zeta?)
 ;; This version is missing normalization conditions 5 and 9, and thus also
 ;; 12 and 8. But it contains the new conditions 13 and 14.
 (let ((j (min j l))
       (k (max k i)))
  (let ((gamma?
	 ;; Gamma is made false when i=-infinity.
	 (and gamma? (not (= i minus-infinity))))
	(delta?
	 ;; Delta is made false when either j=infinity or both j=l and either
	 ;; zeta, alpha, or beta is false.
	 (and delta?
	      (not (= j infinity))
	      (or (not (= j l)) (and zeta? alpha? beta?))))
	(epsilon?
	 ;; Epsilon is made false when either k=-infinity or both k=i and
	 ;; either gamma, beta, or alpha is false.
	 (and epsilon?
	      (not (= k minus-infinity))
	      (or (not (= k i)) (and gamma? beta? alpha?))))
	(zeta?
	 ;; Zeta is made false when i=infinity.
	 (and zeta? (not (= l infinity)))))
   (if (and (<= i j)
	    (<= k l)
	    (implies (= i j) (and gamma? delta?))
	    (implies (= k l) (and epsilon? zeta?))
	    (implies (= i l) (and alpha? beta?))
	    (not (= i infinity))
	    (not (= j minus-infinity))
	    (not (= k infinity))
	    (not (= l minus-infinity)))
       (list (make-extended-spanning-interval
	      i j k l alpha? beta? gamma? delta? epsilon? zeta?))
       '()))))

(define (spanning-intervals i j alpha? beta? gamma? delta?)
 (extended-spanning-intervals
  i j i j alpha? beta? gamma? delta? gamma? delta?))

(define (extended-spanning-interval->string extended-spanning-interval)
 (define (-inf x) (if (= x minus-infinity) "-inf" x))
 (define (inf x) (if (= x infinity) "inf" x))
 (cond
  ((and (= (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-j extended-spanning-interval)
	   (extended-spanning-interval-k extended-spanning-interval)
	   (extended-spanning-interval-l extended-spanning-interval))
	(extended-spanning-interval-alpha? extended-spanning-interval)
	(extended-spanning-interval-beta? extended-spanning-interval)
	(extended-spanning-interval-gamma? extended-spanning-interval)
	(extended-spanning-interval-delta? extended-spanning-interval)
	(extended-spanning-interval-epsilon? extended-spanning-interval)
	(extended-spanning-interval-zeta? extended-spanning-interval))
   (format #f "{{~a}}"
	   (extended-spanning-interval-i extended-spanning-interval)))
  ((and (= (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-j extended-spanning-interval))
	(= (extended-spanning-interval-k extended-spanning-interval)
	   (extended-spanning-interval-l extended-spanning-interval))
	(extended-spanning-interval-gamma? extended-spanning-interval)
	(extended-spanning-interval-delta? extended-spanning-interval)
	(extended-spanning-interval-epsilon? extended-spanning-interval)
	(extended-spanning-interval-zeta? extended-spanning-interval))
   (format #f "~a{~a},{~a}~a"
	   (if (extended-spanning-interval-alpha? extended-spanning-interval)
	       "["
	       "(")
	   (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-k extended-spanning-interval)
	   (if (extended-spanning-interval-beta? extended-spanning-interval)
	       "]"
	       ")")))
  ((and (= (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-j extended-spanning-interval))
	(extended-spanning-interval-gamma? extended-spanning-interval)
	(extended-spanning-interval-delta? extended-spanning-interval))
   (format #f "~a{~a},~a~a,~a~a~a"
	   (if (extended-spanning-interval-alpha? extended-spanning-interval)
	       "["
	       "(")
	   (extended-spanning-interval-i extended-spanning-interval)
	   (if (extended-spanning-interval-epsilon? extended-spanning-interval)
	       "["
	       "(")
	   (-inf (extended-spanning-interval-k extended-spanning-interval))
	   (inf (extended-spanning-interval-l extended-spanning-interval))
	   (if (extended-spanning-interval-zeta? extended-spanning-interval)
	       "]"
	       ")")
	   (if (extended-spanning-interval-beta? extended-spanning-interval)
	       "]"
	       ")")))
  ((and (= (extended-spanning-interval-k extended-spanning-interval)
	   (extended-spanning-interval-l extended-spanning-interval))
	(extended-spanning-interval-epsilon? extended-spanning-interval)
	(extended-spanning-interval-zeta? extended-spanning-interval))
   (format #f "~a~a~a,~a~a,{~a}~a"
	   (if (extended-spanning-interval-alpha? extended-spanning-interval)
	       "["
	       "(")
	   (if (extended-spanning-interval-gamma? extended-spanning-interval)
	       "["
	       "(")
	   (-inf (extended-spanning-interval-i extended-spanning-interval))
	   (inf (extended-spanning-interval-j extended-spanning-interval))
	   (if (extended-spanning-interval-delta? extended-spanning-interval)
	       "]"
	       ")")
	   (extended-spanning-interval-k extended-spanning-interval)
	   (if (extended-spanning-interval-beta? extended-spanning-interval)
	       "]"
	       ")")))
  ;; needs work: This is a special case to undo normalization. It needs to be
  ;;             made more general.
  ((and (= (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-k extended-spanning-interval))
	(= (extended-spanning-interval-j extended-spanning-interval)
	   (extended-spanning-interval-l extended-spanning-interval))
	(extended-spanning-interval-alpha? extended-spanning-interval)
	(not (extended-spanning-interval-beta? extended-spanning-interval))
	(extended-spanning-interval-gamma? extended-spanning-interval)
	(not (extended-spanning-interval-delta? extended-spanning-interval))
	(not (extended-spanning-interval-epsilon? extended-spanning-interval))
	(extended-spanning-interval-zeta? extended-spanning-interval))
   (format #f "[[~a:~a])"
	   (-inf (extended-spanning-interval-i extended-spanning-interval))
	   (inf (extended-spanning-interval-j extended-spanning-interval))))
  ((and (= (extended-spanning-interval-i extended-spanning-interval)
	   (extended-spanning-interval-k extended-spanning-interval))
	(= (extended-spanning-interval-j extended-spanning-interval)
	   (extended-spanning-interval-l extended-spanning-interval))
	(eq? (extended-spanning-interval-gamma? extended-spanning-interval)
	     (extended-spanning-interval-epsilon? extended-spanning-interval))
	(eq? (extended-spanning-interval-delta? extended-spanning-interval)
	     (extended-spanning-interval-zeta? extended-spanning-interval)))
   (format #f "~a~a~a:~a~a~a"
	   (if (extended-spanning-interval-alpha? extended-spanning-interval)
	       "["
	       "(")
	   (if (extended-spanning-interval-gamma? extended-spanning-interval)
	       "["
	       "(")
	   (-inf (extended-spanning-interval-i extended-spanning-interval))
	   (inf (extended-spanning-interval-j extended-spanning-interval))
	   (if (extended-spanning-interval-delta? extended-spanning-interval)
	       "]"
	       ")")
	   (if (extended-spanning-interval-beta? extended-spanning-interval)
	       "]"
	       ")")))
  (else
   (format #f "~a~a~a,~a~a,~a~a,~a~a~a"
	   (if (extended-spanning-interval-alpha? extended-spanning-interval)
	       "["
	       "(")
	   (if (extended-spanning-interval-gamma? extended-spanning-interval)
	       "["
	       "(")
	   (-inf (extended-spanning-interval-i extended-spanning-interval))
	   (inf (extended-spanning-interval-j extended-spanning-interval))
	   (if (extended-spanning-interval-delta? extended-spanning-interval)
	       "]"
	       ")")
	   (if (extended-spanning-interval-epsilon? extended-spanning-interval)
	       "["
	       "(")
	   (-inf (extended-spanning-interval-k extended-spanning-interval))
	   (inf (extended-spanning-interval-l extended-spanning-interval))
	   (if (extended-spanning-interval-zeta? extended-spanning-interval)
	       "]"
	       ")")
	   (if (extended-spanning-interval-beta? extended-spanning-interval)
	       "]"
	       ")")))))

(define (intersection-of is js)
 (reduce
  append
  (map (lambda (i)
	(reduce
	 append
	 (map (lambda (j)
	       (if (and (eq? (extended-spanning-interval-alpha? i)
			     (extended-spanning-interval-alpha? j))
			(eq? (extended-spanning-interval-beta? i)
			     (extended-spanning-interval-beta? j)))
		   (extended-spanning-intervals
		    (max (extended-spanning-interval-i i)
			 (extended-spanning-interval-i j))
		    (min (extended-spanning-interval-j i)
			 (extended-spanning-interval-j j))
		    (max (extended-spanning-interval-k i)
			 (extended-spanning-interval-k j))
		    (min (extended-spanning-interval-l i)
			 (extended-spanning-interval-l j))
		    (extended-spanning-interval-alpha? i)
		    (extended-spanning-interval-beta? i)
		    (cond ((> (extended-spanning-interval-i i)
			      (extended-spanning-interval-i j))
			   (extended-spanning-interval-gamma? i))
			  ((< (extended-spanning-interval-i i)
			      (extended-spanning-interval-i j))
			   (extended-spanning-interval-gamma? j))
			  (else (and (extended-spanning-interval-gamma? i)
				     (extended-spanning-interval-gamma? j))))
		    (cond ((< (extended-spanning-interval-j i)
			      (extended-spanning-interval-j j))
			   (extended-spanning-interval-delta? i))
			  ((> (extended-spanning-interval-j i)
			      (extended-spanning-interval-j j))
			   (extended-spanning-interval-delta? j))
			  (else (and (extended-spanning-interval-delta? i)
				     (extended-spanning-interval-delta? j))))
		    (cond ((> (extended-spanning-interval-k i)
			      (extended-spanning-interval-k j))
			   (extended-spanning-interval-epsilon? i))
			  ((< (extended-spanning-interval-k i)
			      (extended-spanning-interval-k j))
			   (extended-spanning-interval-epsilon? j))
			  (else (and (extended-spanning-interval-epsilon? i)
				     (extended-spanning-interval-epsilon? j))))
		    (cond ((< (extended-spanning-interval-l i)
			      (extended-spanning-interval-l j))
			   (extended-spanning-interval-zeta? i))
			  ((> (extended-spanning-interval-l i)
			      (extended-spanning-interval-l j))
			   (extended-spanning-interval-zeta? j))
			  (else (and (extended-spanning-interval-zeta? i)
				     (extended-spanning-interval-zeta? j)))))
		   '()))
	      js)
	 '()))
       is)
  '()))

(define (complement-of is)
 (if (null? is)
     (all-intervals)
     (intersection-of
      (let ((i (first is)))
       (append (extended-spanning-intervals
		minus-infinity
		infinity
		minus-infinity
		(extended-spanning-interval-k i)
		(extended-spanning-interval-alpha? i)
		(extended-spanning-interval-beta? i)
		#t
		#t
		#t
		(not (extended-spanning-interval-epsilon? i)))
	       (extended-spanning-intervals
		minus-infinity
		infinity
		(extended-spanning-interval-l i)
		infinity
		(extended-spanning-interval-alpha? i)
		(extended-spanning-interval-beta? i)
		#t
		#t
		(not (extended-spanning-interval-zeta? i))
		#t)
	       (extended-spanning-intervals
		minus-infinity
		(extended-spanning-interval-i i)
		minus-infinity
		infinity
		(extended-spanning-interval-alpha? i)
		(extended-spanning-interval-beta? i)
		#t
		(not (extended-spanning-interval-gamma? i))
		#t
		#t)
	       (extended-spanning-intervals
		(extended-spanning-interval-j i)
		infinity
		minus-infinity
		infinity
		(extended-spanning-interval-alpha? i)
		(extended-spanning-interval-beta? i)
		(not (extended-spanning-interval-delta? i))
		#t
		#t
		#t)
	       (extended-spanning-intervals
		minus-infinity
		infinity
		minus-infinity
		infinity
		(not (extended-spanning-interval-alpha? i))
		(extended-spanning-interval-beta? i)
		#t
		#t
		#t
		#t)
	       (extended-spanning-intervals
		minus-infinity
		infinity
		minus-infinity
		infinity
		(extended-spanning-interval-alpha? i)
		(not (extended-spanning-interval-beta? i))
		#t
		#t
		#t
		#t)
	       (extended-spanning-intervals
		minus-infinity
		infinity
		minus-infinity
		infinity
		(not (extended-spanning-interval-alpha? i))
		(not (extended-spanning-interval-beta? i))
		#t
		#t
		#t
		#t)))
      (complement-of (rest is)))))

(define (resolution-of i j)
 (append (if (and (eq? (extended-spanning-interval-alpha? i)
		       (extended-spanning-interval-alpha? j))
		  (eq? (extended-spanning-interval-beta? i)
		       (extended-spanning-interval-beta? j))
		  (<= (extended-spanning-interval-i i)
		      (extended-spanning-interval-j j))
		  (<= (extended-spanning-interval-i j)
		      (extended-spanning-interval-j i))
		  (implies (or (not (extended-spanning-interval-gamma? i))
			       (not (extended-spanning-interval-delta? j)))
			   (not (= (extended-spanning-interval-i i)
				   (extended-spanning-interval-j j))))
		  (implies (or (not (extended-spanning-interval-gamma? j))
			       (not (extended-spanning-interval-delta? i)))
			   (not (= (extended-spanning-interval-i j)
				   (extended-spanning-interval-j i)))))
	     (extended-spanning-intervals
	      (min (extended-spanning-interval-i i)
		   (extended-spanning-interval-i j))
	      (max (extended-spanning-interval-j i)
		   (extended-spanning-interval-j j))
	      (max (extended-spanning-interval-k i)
		   (extended-spanning-interval-k j))
	      (min (extended-spanning-interval-l i)
		   (extended-spanning-interval-l j))
	      (extended-spanning-interval-alpha? i)
	      (extended-spanning-interval-beta? i)
	      (cond ((< (extended-spanning-interval-i i)
			(extended-spanning-interval-i j))
		     (extended-spanning-interval-gamma? i))
		    ((> (extended-spanning-interval-i i)
			(extended-spanning-interval-i j))
		     (extended-spanning-interval-gamma? j))
		    (else (or (extended-spanning-interval-gamma? i)
			      (extended-spanning-interval-gamma? j))))
	      (cond ((> (extended-spanning-interval-j i)
			(extended-spanning-interval-j j))
		     (extended-spanning-interval-delta? i))
		    ((< (extended-spanning-interval-j i)
			(extended-spanning-interval-j j))
		     (extended-spanning-interval-delta? j))
		    (else (or (extended-spanning-interval-delta? i)
			      (extended-spanning-interval-delta? j))))
	      (cond ((> (extended-spanning-interval-k i)
			(extended-spanning-interval-k j))
		     (extended-spanning-interval-epsilon? i))
		    ((< (extended-spanning-interval-k i)
			(extended-spanning-interval-k j))
		     (extended-spanning-interval-epsilon? j))
		    (else (and (extended-spanning-interval-epsilon? i)
			       (extended-spanning-interval-epsilon? j))))
	      (cond ((< (extended-spanning-interval-l i)
			(extended-spanning-interval-l j))
		     (extended-spanning-interval-zeta? i))
		    ((> (extended-spanning-interval-l i)
			(extended-spanning-interval-l j))
		     (extended-spanning-interval-zeta? j))
		    (else (and (extended-spanning-interval-zeta? i)
			       (extended-spanning-interval-zeta? j)))))
	     '())
	 (if (and (eq? (extended-spanning-interval-alpha? i)
		       (extended-spanning-interval-alpha? j))
		  (eq? (extended-spanning-interval-beta? i)
		       (extended-spanning-interval-beta? j))
		  (<= (extended-spanning-interval-k i)
		      (extended-spanning-interval-l j))
		  (<= (extended-spanning-interval-k j)
		      (extended-spanning-interval-l i))
		  (implies (or (not (extended-spanning-interval-epsilon? i))
			       (not (extended-spanning-interval-zeta? j)))
			   (not (= (extended-spanning-interval-k i)
				   (extended-spanning-interval-l j))))
		  (implies (or (not (extended-spanning-interval-epsilon? j))
			       (not (extended-spanning-interval-zeta? i)))
			   (not (= (extended-spanning-interval-k j)
				   (extended-spanning-interval-l i)))))
	     (extended-spanning-intervals
	      (max (extended-spanning-interval-i i)
		   (extended-spanning-interval-i j))
	      (min (extended-spanning-interval-j i)
		   (extended-spanning-interval-j j))
	      (min (extended-spanning-interval-k i)
		   (extended-spanning-interval-k j))
	      (max (extended-spanning-interval-l i)
		   (extended-spanning-interval-l j))
	      (extended-spanning-interval-alpha? i)
	      (extended-spanning-interval-beta? i)
	      (cond ((> (extended-spanning-interval-i i)
			(extended-spanning-interval-i j))
		     (extended-spanning-interval-gamma? i))
		    ((< (extended-spanning-interval-i i)
			(extended-spanning-interval-i j))
		     (extended-spanning-interval-gamma? j))
		    (else (and (extended-spanning-interval-gamma? i)
			       (extended-spanning-interval-gamma? j))))
	      (cond ((< (extended-spanning-interval-j i)
			(extended-spanning-interval-j j))
		     (extended-spanning-interval-delta? i))
		    ((> (extended-spanning-interval-j i)
			(extended-spanning-interval-j j))
		     (extended-spanning-interval-delta? j))
		    (else (and (extended-spanning-interval-delta? i)
			       (extended-spanning-interval-delta? j))))
	      (cond ((< (extended-spanning-interval-k i)
			(extended-spanning-interval-k j))
		     (extended-spanning-interval-epsilon? i))
		    ((> (extended-spanning-interval-k i)
			(extended-spanning-interval-k j))
		     (extended-spanning-interval-epsilon? j))
		    (else (or (extended-spanning-interval-epsilon? i)
			      (extended-spanning-interval-epsilon? j))))
	      (cond ((> (extended-spanning-interval-l i)
			(extended-spanning-interval-l j))
		     (extended-spanning-interval-zeta? i))
		    ((< (extended-spanning-interval-l i)
			(extended-spanning-interval-l j))
		     (extended-spanning-interval-zeta? j))
		    (else (or (extended-spanning-interval-zeta? i)
			      (extended-spanning-interval-zeta? j)))))
	     '())))

(define (subsumes? j i)
 (and (eq? (extended-spanning-interval-alpha? i)
	   (extended-spanning-interval-alpha? j))
      (eq? (extended-spanning-interval-beta? i)
	   (extended-spanning-interval-beta? j))
      (>= (extended-spanning-interval-i i)
	  (extended-spanning-interval-i j))
      (<= (extended-spanning-interval-j i)
	  (extended-spanning-interval-j j))
      (>= (extended-spanning-interval-k i)
	  (extended-spanning-interval-k j))
      (<= (extended-spanning-interval-l i)
	  (extended-spanning-interval-l j))
      (implies (and (extended-spanning-interval-gamma? i)
		    (not (extended-spanning-interval-gamma? j)))
	       (not (= (extended-spanning-interval-i i)
		       (extended-spanning-interval-i j))))
      (implies (and (extended-spanning-interval-delta? i)
		    (not (extended-spanning-interval-delta? j)))
	       (not (= (extended-spanning-interval-j i)
		       (extended-spanning-interval-j j))))
      (implies (and (extended-spanning-interval-epsilon? i)
		    (not (extended-spanning-interval-epsilon? j)))
	       (not (= (extended-spanning-interval-k i)
		       (extended-spanning-interval-k j))))
      (implies (and (extended-spanning-interval-zeta? i)
		    (not (extended-spanning-interval-zeta? j)))
	       (not (= (extended-spanning-interval-l i)
		       (extended-spanning-interval-l j))))))

(define (span is js)
 (reduce
  append
  (map
   (lambda (i)
    (reduce
     append
     (map
      (lambda (j)
       (let* ((i1 (extended-spanning-interval-i i))
	      (j1 (extended-spanning-interval-j i))
	      (k1 (extended-spanning-interval-k i))
	      (l1 (extended-spanning-interval-l i))
	      (alpha1? (extended-spanning-interval-alpha? i))
	      (beta1? (extended-spanning-interval-beta? i))
	      (gamma1? (extended-spanning-interval-gamma? i))
	      (delta1? (extended-spanning-interval-delta? i))
	      (epsilon1? (extended-spanning-interval-epsilon? i))
	      (zeta1? (extended-spanning-interval-zeta? i))
	      (i2 (extended-spanning-interval-i j))
	      (j2 (extended-spanning-interval-j j))
	      (k2 (extended-spanning-interval-k j))
	      (l2 (extended-spanning-interval-l j))
	      (alpha2? (extended-spanning-interval-alpha? j))
	      (beta2? (extended-spanning-interval-beta? j))
	      (gamma2? (extended-spanning-interval-gamma? j))
	      (delta2? (extended-spanning-interval-delta? j))
	      (epsilon2? (extended-spanning-interval-epsilon? j))
	      (zeta2? (extended-spanning-interval-zeta? j))
	      (j-prime (min j1 j2))
	      (k-prime (max k1 k2))
	      (delta?-prime (or (and delta1? (< j1 j2))
				(and delta2? (> j1 j2))
				(and delta1? delta2? (= j1 j2))))
	      (epsilon?-prime (or (and epsilon1? (> k1 k2))
				  (and epsilon2? (< k1 k2))
				  (and epsilon1? epsilon2? (= k1 k2)))))
	(append (extended-spanning-intervals
		 i1 j-prime k-prime l1
		 alpha1? beta1? gamma1? delta?-prime epsilon?-prime zeta1?)
		(extended-spanning-intervals
		 i1 j-prime k-prime l2
		 alpha1? beta2? gamma1? delta?-prime epsilon?-prime zeta2?)
		(extended-spanning-intervals
		 i2 j-prime k-prime l1
		 alpha2? beta1? gamma2? delta?-prime epsilon?-prime zeta1?)
		(extended-spanning-intervals
		 i2 j-prime k-prime l2
		 alpha2? beta2? gamma2? delta?-prime epsilon?-prime zeta2?))))
      js)
     '()))
   is)
  '()))

(define (diamond-allen r is)
 (reduce
  append
  (map
   (lambda (i)
    (let ((i1 (extended-spanning-interval-i i))
	  (j1 (extended-spanning-interval-j i))
	  (k1 (extended-spanning-interval-k i))
	  (l1 (extended-spanning-interval-l i))
	  (alpha1? (extended-spanning-interval-alpha? i))
	  (beta1? (extended-spanning-interval-beta? i))
	  (gamma1? (extended-spanning-interval-gamma? i))
	  (delta1? (extended-spanning-interval-delta? i))
	  (epsilon1? (extended-spanning-interval-epsilon? i))
	  (zeta1? (extended-spanning-interval-zeta? i)))
     (case r
      ((=) (list i))
      ((<) (append (extended-spanning-intervals
		    (- k1 *fuzz*) infinity minus-infinity infinity
		    #f #f (and (not beta1?) epsilon1?) #t #t #t)
		   (extended-spanning-intervals
		    (- k1 *fuzz*) infinity minus-infinity infinity
		    #f #t (and (not beta1?) epsilon1?) #t #t #t)
		   (extended-spanning-intervals
		    (- k1 *fuzz*) infinity minus-infinity infinity
		    #t #f #f #t #t #t)
		   (extended-spanning-intervals
		    (- k1 *fuzz*) infinity minus-infinity infinity
		    #t #t #f #t #t #t)))
      ((>) (append (extended-spanning-intervals
		    minus-infinity infinity minus-infinity (+ j1 *fuzz*)
		    #f #f #t #t #t (and (not alpha1?) delta1?))
		   (extended-spanning-intervals
		    minus-infinity infinity minus-infinity (+ j1 *fuzz*)
		    #f #t #t #t #t #f)
		   (extended-spanning-intervals
		    minus-infinity infinity minus-infinity (+ j1 *fuzz*)
		    #t #f #t #t #t (and (not alpha1?) delta1?))
		   (extended-spanning-intervals
		    minus-infinity infinity minus-infinity (+ j1 *fuzz*)
		    #t #t #t #t #t #f)))
      ((m) (append (extended-spanning-intervals
		    (- k1 *fuzz*) (+ l1 *fuzz*) minus-infinity infinity
		    (not beta1?) #f epsilon1? zeta1? #t #t)
		   (extended-spanning-intervals
		    (- k1 *fuzz*) (+ l1 *fuzz*)  minus-infinity infinity
		    (not beta1?) #t epsilon1? zeta1? #t #t)))
      ((mi) (append (extended-spanning-intervals
		     minus-infinity infinity (- i1 *fuzz*) (+ j1 *fuzz*)
		     #f (not alpha1?) #t #t gamma1? delta1?)
		    (extended-spanning-intervals
		     minus-infinity infinity (- i1 *fuzz*) (+ j1 *fuzz*)
		     #t (not alpha1?) #t #t gamma1? delta1?)))
      ((o) (append
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ l1 *fuzz*) (- k1 *fuzz*) infinity
	     #f #f (and alpha1? gamma1?) #f #f #t)
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ l1 *fuzz*) (- k1 *fuzz*) infinity
	     #t #f #f (and beta1? zeta1?) #f #t)
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ l1 *fuzz*) (- k1 *fuzz*) infinity
	     #f #t (and alpha1? gamma1?) #f (and (not beta1?) epsilon1?) #t)
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ l1 *fuzz*) (- k1 *fuzz*) infinity
	     #t #t #f (and beta1? zeta1?) (and (not beta1?) epsilon1?) #t)))
      ((oi) (append
	     (extended-spanning-intervals
	      minus-infinity (+ j1 *fuzz*) (- i1 *fuzz*) (+ l1 *fuzz*)
	      #f #f #t #f #f (and beta1? zeta1?))
	     (extended-spanning-intervals
	      minus-infinity (+ j1 *fuzz*) (- i1 *fuzz*) (+ l1 *fuzz*)
	      #t #f #t (and (not alpha1?) delta1?) #f (and beta1? zeta1?))
	     (extended-spanning-intervals
	      minus-infinity (+ j1 *fuzz*) (- i1 *fuzz*) (+ l1 *fuzz*)
	      #f #t #t #f (and alpha1? gamma1?) #f)
	     (extended-spanning-intervals
	      minus-infinity (+ j1 *fuzz*) (- i1 *fuzz*) (+ l1 *fuzz*)
	      #t #t #t (and (not alpha1?) delta1?) (and alpha1? gamma1?) #f)))
      ((s) (append
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ j1 *fuzz*) (- k1 *fuzz*) infinity
	     alpha1? #f gamma1? delta1? #f #t)
	    (extended-spanning-intervals
	     (- i1 *fuzz*) (+ j1 *fuzz*) (- k1 *fuzz*) infinity
	     alpha1? #t gamma1? delta1? (and (not beta1?) epsilon1?) #t)))
      ((si) (append (extended-spanning-intervals
		     (- i1 *fuzz*) (+ j1 *fuzz*) minus-infinity (+ l1 *fuzz*)
		     alpha1? #f gamma1? delta1? #t (and beta1? zeta1?))
		    (extended-spanning-intervals
		     (- i1 *fuzz*) (+ j1 *fuzz*) minus-infinity (+ l1 *fuzz*)
		     alpha1? #t gamma1? delta1? #t #f)))
      ((f) (append
	    (extended-spanning-intervals
	     minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) (+ l1 *fuzz*)
	     #f beta1? #t #f epsilon1? zeta1?)
	    (extended-spanning-intervals
	     minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) (+ l1 *fuzz*)
	     #t beta1? #t (and (not alpha1?) delta1?) epsilon1? zeta1?)))
      ((fi) (append (extended-spanning-intervals
		     (- i1 *fuzz*) infinity (- k1 *fuzz*) (+ l1 *fuzz*)
		     #f beta1? (and alpha1? gamma1?) #t epsilon1? zeta1?)
		    (extended-spanning-intervals
		     (- i1 *fuzz*) infinity (- k1 *fuzz*) (+ l1 *fuzz*)
		     #t beta1? #f #t epsilon1? zeta1?)))
      ((d) (append (extended-spanning-intervals
		    minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) infinity
		    #f #f #t #f #f #t)
		   (extended-spanning-intervals
		    minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) infinity
		    #t #f #t (and (not alpha1?) delta1?) #f #t)
		   (extended-spanning-intervals
		    minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) infinity
		    #f #t #t #f (and (not beta1?) epsilon1?) #t)
		   (extended-spanning-intervals
		    minus-infinity (+ j1 *fuzz*) (- k1 *fuzz*) infinity
		    #t
		    #t
		    #t
		    (and (not alpha1?) delta1?)
		    (and (not beta1?) epsilon1?)
		    #t)))
      ((di) (append (extended-spanning-intervals
		     (- i1 *fuzz*) infinity minus-infinity (+ l1 *fuzz*)
		     #f #f (and alpha1? gamma1?) #t #t (and beta1? zeta1?))
		    (extended-spanning-intervals
		     (- i1 *fuzz*) infinity minus-infinity (+ l1 *fuzz*)
		     #f #t (and alpha1? gamma1?) #t #t #f)
		    (extended-spanning-intervals
		     (- i1 *fuzz*) infinity minus-infinity (+ l1 *fuzz*)
		     #t #f #f #t #t (and beta1? zeta1?))
		    (extended-spanning-intervals
		     (- i1 *fuzz*) infinity minus-infinity (+ l1 *fuzz*)
		     #t #t #f #t #t #f)))
      (else (fuck-up)))))
   is)
  '()))

(define (invert-allen r)
 (case r
  ((=) '=)
  ((<) '>)
  ((>) '<)
  ((m) 'mi)
  ((mi) 'm)
  ((o) 'oi)
  ((oi) 'o)
  ((s) 'si)
  ((si) 's)
  ((f) 'fi)
  ((fi) 'f)
  ((d) 'di)
  ((di) 'd)
  (else (fuck-up))))

(define (span-allen js r ks)
 (reduce
  append
  (map
   (lambda (j)
    (reduce
     append
     (map
      (lambda (k)
       (span
	(intersection-of (diamond-allen r (list j)) (list k))
	(intersection-of (diamond-allen (invert-allen r) (list k)) (list j))))
      ks)
     '()))
   js)
  '()))

(define (close-under-resolution-and-subsumption
	 resolution-function subsumption-predicate set)
 (let ((values '()))
  (let loop ()
   (unless (null? set)
    (let ((new-value (first set)))
     (set! set (rest set))
     (unless (some (lambda (old-value)
		    (subsumption-predicate old-value new-value))
		   values)
      (set! values
	    (remove-if
	     (lambda (old-value) (subsumption-predicate new-value old-value))
	     values))
      (for-each
       (lambda (old-value)
	(set! set (append (resolution-function new-value old-value) set)))
       values)
      (set! values (cons new-value values))))
    (loop)))
  values))

(define (make-irredundant is)
 (close-under-resolution-and-subsumption resolution-of subsumes? is))

